#! /usr/bin/env python
#
# Test(s) for ../text_categorizer.py
#
# Notes:
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_<module>.py
#

"""Tests for text_categorizer module"""

# Standard packages
import re

# Installed packages
import pytest

# Local packages
from mezcla.unittest_wrapper import TestWrapper
from mezcla import debug
## TODO: from mezcla import system
from mezcla import glue_helpers as gh

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
import mezcla.text_categorizer as THE_MODULE
#
# Note: sanity test for customization (TODO: remove if desired)
if not re.search(__file__, r"\btemplate.py$"):
    debug.assertion("mezcla.template" not in str(THE_MODULE))

## TODO:
## # Environment options
## # Note: These are just intended for internal options, not for end users.
## # It also allows for enabling options in one place.
## #
## FUBAR = system.getenv_bool("FUBAR", False,
##                            description="Fouled Up Beyond All Recognition processing")


class TestTextCategorizerUtils(TestWrapper):
    """Class for testcase definition"""

    def test_something_else(self):
        """TODO: flesh out test for something else"""
        debug.trace(4, "test_something_else()")
        ## ex: assert THE_MODULE.TODO_function() == TODO_value

    def test_sklearn_report(self):
        """Ensure sklearn_report works as expected"""
        debug.trace(4, "test_sklearn_report()")
        ## TODO: WORK-IN=PROGRESS

    def test_create_tabular_file(self):
        """Ensure create_tabular_file works as expected"""
        debug.trace(4, "test_create_tabular_file()")
        ## TODO: WORK-IN=PROGRESS

    def test_read_categorization_data(self):
        """Ensure read_categorization_data works as expected"""
        debug.trace(4, "test_read_categorization_data()")
        ## TODO: WORK-IN=PROGRESS

    def test_int_if_whole(self):
        """Ensure int_if_whole works as expected"""
        debug.trace(4, "test_int_if_whole()")
        assert THE_MODULE.int_if_whole(2.0) == 2
        assert THE_MODULE.int_if_whole(2.999) == 2.999

    def test_param_or_default(self):
        """Ensure param_or_default works as expected"""
        debug.trace(4, "test_param_or_default()")
        ## TODO: WORK-IN=PROGRESS

    @pytest.mark.skip(reason="TODO: test not yet implemented")
    def test_data_file(self):
        """Makes sure TODO works as expected"""
        debug.trace(4, "TestTextCategorizer.test_data_file()")
        data = ["TODO1", "TODO2"]
        gh.write_lines(self.temp_file, data)
        output = self.run_script("", self.temp_file)
        assert re.search(r"TODO-pattern", output.strip())
        return

    def test_format_index_html(self):
        """Ensure format_index_html works as expected"""
        debug.trace(4, "test_format_index_html()")
        ## TODO: WORK-IN=PROGRESS

    def test_start_web_controller(self):
        """Ensure start_web_controller works as expected"""
        debug.trace(4, "test_start_web_controller()")
        ## TODO: WORK-IN=PROGRESS


class TestTextCategorizerScript(TestWrapper):
    """Class for testcase definition"""
    script_file = TestWrapper.get_module_file_path(__file__)
    script_module = TestWrapper.get_testing_module_name(__file__)
    # TODO: use_temp_base_dir = True            # treat TEMP_BASE as directory
    # note: temp_file defined by parent (along with script_module, temp_base, and test_num)

    ## TODO: WORK-IN-PROGRESS TESTS


#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    pytest.main([__file__])



#! /usr/bin/env python
# 
# Sample script using Main class. By default, this outputs lines that contain
# "fubar". There is an option to check for lines matching a regular expression.
#

"""Simple illustration of Main class"""

# Standard packages
import re

# Local packages
from mezcla import debug
from mezcla.main import Main

class Script(Main):
    """Input processing class"""
    regex = None
    check_fubar = None

    def setup(self):
        """Check results of command line processing"""
        debug.trace_fmtd(5, "Script.setup(): self={s}", s=self)
        self.regex = self.parsed_args['regex']
        self.check_fubar = self.get_parsed_option('check-fubar', not self.regex)
        debug.assertion(bool(self.regex) ^ self.check_fubar)
        debug.trace_object(5, self, label=f"{self.__class__.__name__} instance")

    def process_line(self, line):
        """Processes current line from input"""
        debug.trace_fmtd(6, "Script.process_line({l})", l=line)
        if self.check_fubar and "fubar" in line:
            debug.trace(4, f"Fubar line {self.line_num}: {line}")
            print(line)
        elif self.regex and re.search(self.regex, line):
            debug.trace(4, f"Regex line {self.line_num}: {line}")
            print(line)
        return

if __name__ == '__main__':
    ## TODO: app => script???
    app = Script(description=__doc__,
                 boolean_options=["check-fubar"],
                 text_options=[("regex", "Regular expression")])
    app.run()



#! /usr/bin/env python
#
# Illustrates how to use Stable Diffusion via Hugging Face diffuser package,
# including gradio-based UI.
#
# via https://huggingface.co/spaces/stabilityai/stable-diffusion
# also uses https://huggingface.co/CompVis/stable-diffusion-v1-4
# and https://stackoverflow.com/questions/48273205/accessing-incoming-post-data-in-flask
#
# Note:
# - For tips on parameter settings, see
#   https://getimg.ai/guides/interactive-guide-to-stable-diffusion-guidance-scale-parameter
#
# TODO:
# - Set GRADIO_SERVER_NAME to 0.0.0.0?
#

"""Image generation via HF Stable Diffusion (SD) API"""

# Standard modules
import base64
import json
import time

# Installed modules
import diskcache
from flask import Flask, request
import PIL
import requests
import gradio as gr

# Local modules
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.main import Main
from mezcla.my_regex import my_re
from mezcla import system

# Constants/globals
TL = debug.TL
PROMPT = system.getenv_text("PROMPT", "your favorite politician in a tutu",
                            "Textual prompt describing image")
NEGATIVE_PROMPT = system.getenv_text("NEGATIVE_PROMPT", "photo realistic",
                            "Negative tips for image")
GUIDANCE_SCALE = system.getenv_int("GUIDANCE_SCALE", 7,
                                   "How much the image generation follows the prompt")
SD_URL = system.getenv_value("SD_URL", None,
                             "URL for SD TCP/restful serve--new via flask or remote")
SD_PORT = system.getenv_int("SD_PORT", 9700,
                            "TCP port for SD server")
SD_DEBUG = system.getenv_int("SD_DEBUG", False,
                             "Run SD server in debug mode")
USE_HF_API = system.getenv_bool("USE_HF_API", not SD_URL,
                                "Use Huggingface API instead of TCP server")
CHECK_UNSAFE = system.getenv_bool("CHECK_UNSAFE", False,
                                  "Apply unsafe word list regex filter")
NUM_IMAGES = system.getenv_int("NUM_IMAGES", 1,
                               "Number of images to generated")
BASENAME = system.getenv_text("BASENAME", "sd-app-image",
                              "Basename for saving images")
LOW_MEMORY = system.getenv_bool("LOW_MEMORY", False,
                                "Use low memory computations such as via float16")
DUMMY_RESULT = system.getenv_bool("DUMMY_RESULT", False,
                                  "Mock up SD server result")
DISK_CACHE = system.getenv_value("SD_DISK_CACHE", None,
                                 "Path to directory with disk cache")

BATCH_ARG = "batch"
SERVER_ARG = "server"
UI_ARG = "UI"
PORT_ARG = "port"
PROMPT_ARG = "prompt"
NEGATIVE_ARG = "negative"
GUIDANCE_ARG = "guidance"

# Conditional imports for HG/PyTorch
torch = None
load_dataset = None
if USE_HF_API:
    # pylint: disable=import-outside-toplevel, import-error
    from datasets import load_dataset
    import torch

word_list = []
if CHECK_UNSAFE:
    word_list_dataset = load_dataset("stabilityai/word-list", data_files="list.txt", use_auth_token=True)
    word_list = word_list_dataset["train"]['text']
    debug.trace_expr(5, word_list)

sd_instance = None
flask_app = Flask(__name__)


def show_gpu_usage(level=TL.DETAILED):
    """Show usage for GPU memory, etc.
    TODO: support other types besides NVidia"""
    debug.trace(level, "GPU usage")
    debug.trace(level, gh.run("nvidia-smi"))

#-------------------------------------------------------------------------------
# Main Stable Diffusion support

class StableDiffusion:
    """Class providing Stable Diffusion generative AI (e.g., text-to-image)"""

    def __init__(self, use_hf_api=None, server_url=None, server_port=None, low_memory=None):
        debug.trace(4, f"{self.__class__.__name__}.__init__{(use_hf_api, server_url, server_port)}")
        if use_hf_api is None:
            use_hf_api = USE_HF_API
        self.use_hf_api = use_hf_api
        if (server_url is None) and (not self.use_hf_api):
            server_url = SD_URL
        self.server_url = server_url
        if server_port is None:
            server_port = SD_PORT
        if self.server_url and not my_re.search(r"^https?", self.server_url):
            # note: remote flask server (e.g., on GPU server)
            self.server_url = f"http://{self.server_url}"
            debug.trace(4, f"Added http protocol to URL: {self.server_url}")
        if self.server_url and not my_re.search(r":\d+", self.server_url):
            # TODO3: http://base-url/path => http://base-url:port/path
            self.server_url += f":{server_port}"
        else:
            system.print_stderr(f"Warning: ignoring port {server_port} as already in URL {self.server_url}")
        if low_memory is None:
            low_memory = LOW_MEMORY
        self.low_memory = low_memory
        self.pipe = None
        self.cache = None
        if DISK_CACHE:
            self.cache = diskcache.Cache(
                DISK_CACHE,                   # path to dir
                disk=diskcache.core.JSONDisk, # avoid serialization issue
                disk_compress_level=0,        # no compression
                cull_limit=0)                 # no automatic pruning
        debug.assertion(bool(self.use_hf_api) != bool(self.server_url))
        debug.trace_object(5, self, label=f"{self.__class__.__name__} instance")
    
    def init_pipeline(self):
        """Initialize Stable Diffusion"""
        debug.trace(4, "init_pipeline()")
        # pylint: disable=import-outside-toplevel
        from diffusers import StableDiffusionPipeline
        model_id = "CompVis/stable-diffusion-v1-4"
        device = "cuda"
        # TODO2: automatically use LOW_MEMORY if GPU memory below 8gb
        dtype=(torch.float16 if self.low_memory else None)
        pipe = StableDiffusionPipeline.from_pretrained(model_id, torch_dtype=dtype)
        debug.trace_expr(5, pipe, dtype)
        pipe = pipe.to(device)
        pipe.set_progress_bar_config(disable=True)
        pipe.enable_attention_slicing()
        debug.trace_object(6, pipe)
        show_gpu_usage()
        return pipe


    def infer(self, prompt=None, negative_prompt=None, scale=None, num_images=None,
              skip_img_spec=False):
        """Generate images using positive PROMPT and NEGATIVE one, along with guidance SCALE
        Returns list of NUM image specifications in base64 format (e.g., for use in HTML)
        Note: If SKIP_IMG_SPEC specified, result is formatted for HTML IMG tag
        """
        ## OLD: debug.trace(4, f"{self.__class__.__name__}.infer{(prompt, negative_prompt, scale, num_images)}")
        debug.trace_expr(4, prompt, negative_prompt, scale, num_images, skip_img_spec, prefix=f"in {self.__class__.__name__}.infer:\n\t", delim="\n\t", max_len=1024)
        if num_images is None:
            num_images = NUM_IMAGES
        if scale is None:
            scale = GUIDANCE_SCALE
        for prompt_filter in word_list:
            if my_re.search(rf"\b{prompt_filter}\b", prompt):
                raise gr.Error("Unsafe content found. Please try again with different prompts.")
    
        images = []
        params = (prompt, negative_prompt, scale, num_images, skip_img_spec)

        if self.cache is not None:
            images = self.cache.get(params)
        if images and len(images) > 0:
            ## TEST:
            ## debug.trace_fmt(6, "Using cached result for params {p}: ({r})",
            ##                 p=params, r=images)
            debug.trace_fmt(5, "Using cached infer result: ({r})", r=images)
        else:
            images = self.infer_non_cached(prompt, negative_prompt, scale, num_images, skip_img_spec)
            if self.cache is not None:
                self.cache.set(params, images)
                debug.trace_fmt(6, "Setting cached result (r={r})", r=images)
        return images
            
    def infer_non_cached(self, prompt, negative_prompt, scale, num_images, skip_img_spec):
        """Non-cached version of infer"""
        debug.trace(5, f"{self.__class__.__name__}.infer_non_cached{(prompt, negative_prompt, scale, num_images)}")
        images = []
        if self.use_hf_api:
            if DUMMY_RESULT:
                result = ["iVBORw0KGgoAAAANSUhEUgAAAA8AAAAPAgMAAABGuH3ZAAAADFBMVEUAAMzMzP////8AAABGA1scAAAAJUlEQVR4nGNgAAFGQUEowRoa6sCABBZowAgsgBEIGUQCRALAPACMHAOQvR4HGwAAAABJRU5ErkJggg=="]
                debug.trace(5, f"early exit infer_non_cached() => {result}")
                return result

            if not self.pipe:
                self.pipe = self.init_pipeline()
            start_time = time.time()
            image_info = self.pipe(prompt, negative_prompt=negative_prompt, guidance_scale=scale,
                                   num_images_per_prompt=num_images)
            debug.trace_expr(4, image_info)
            debug.trace_object(5, image_info, "image_info")
            num_generated = 0
            for i, image in enumerate(image_info.images):
                debug.trace_expr(4, image)
                debug.trace_object(5, image, "image")
                b64_encoding = image
                debug.assertion(isinstance(image, PIL.Image.Image))
                try:
                    image_path = f"{BASENAME}-{i + 1}.png"
                    image.save(image_path)
                    num_generated += 1
                    # note: "decodes" base-64 encoded bytes object into UTF-8 string
                    b64_encoding = (base64.b64encode(system.read_binary_file(image_path))).decode()
                except:
                    system.print_exception_info("image-to-base64")
                images.append(b64_encoding)
            elapsed = round(time.time() - start_time, 3)
            debug.trace(4, f"{elapsed} seconds to generate {num_images} images")
            debug.assertion(num_generated == num_images)
            show_gpu_usage()
        else:
            debug.assertion(self.server_url)
            url = self.server_url
            payload = {'prompt': prompt, 'negative_prompt': negative_prompt, 'scale': scale,
                       'num_images': num_images}
            images_request = requests.post(url, json=payload, timeout=(5 * 60))
            debug.trace_object(6, images_request)
            debug.trace_expr(5, payload, images_request, images_request.json(), delim="\n")
            for image in images_request.json()["images"]:
                image_b64 = image
                if not skip_img_spec:
                    image_b64 = (f"data:image/png;base64,{image_b64}")
                images.append(image_b64)
        result = images
        debug.trace_fmt(5, "infer_non_cached() => {r}", r=result)
        
        return result

#-------------------------------------------------------------------------------
# Middleware

@flask_app.route('/', methods=['GET', 'POST'])
def handle_infer():
    """Process request to do inference to generate image from text"""
    debug.trace(6, "[flask_app /] handle_infer()")
    # TODO3: request => flask_request
    debug.trace_object(5, request)
    params = request.get_json()
    debug.trace_expr(5, params)
    images_spec = {"images": sd_instance.infer(**params)}
    # note: see https://stackoverflow.com/questions/45412228/sending-json-and-status-code-with-a-flask-response
    result = (json.dumps(images_spec), 200)
    debug.trace_object(7, result)
    debug.trace_fmt(7, "handle_infer() => {r}", r=result)
    return result


def infer(prompt=None, negative_prompt=None, scale=None, num_images=None, skip_img_spec=None):
    """Wrapper around StableDiffusion.infer()
    Note: intended just for the gradio UI"
    """
    debug.trace(6, f"[sd_instance] infer{(prompt, negative_prompt, scale, skip_img_spec)}")
    return sd_instance.infer(prompt=prompt, negative_prompt=negative_prompt, scale=scale, num_images=num_images, skip_img_spec=skip_img_spec)

#-------------------------------------------------------------------------------
# User interface

def run_ui():
    """Run user interface via gradio serving by default at localhost:7860
    Note: The environment variable GRADIO_SERVER_NAME can be used to serve via 0.0.0.0"""
    css = """
            .gradio-container {
                font-family: 'IBM Plex Sans', sans-serif;
            }
            .gr-button {
                color: white;
                border-color: black;
                background: black;
            }
            input[type='range'] {
                accent-color: black;
            }
            .dark input[type='range'] {
                accent-color: #dfdfdf;
            }
            .container {
                max-width: 730px;
                margin: auto;
                padding-top: 1.5rem;
            }
            #gallery {
                min-height: 22rem;
                margin-bottom: 15px;
                margin-left: auto;
                margin-right: auto;
                border-bottom-right-radius: .5rem !important;
                border-bottom-left-radius: .5rem !important;
            }
            #gallery>div>.h-full {
                min-height: 20rem;
            }
            .details:hover {
                text-decoration: underline;
            }
            .gr-button {
                white-space: nowrap;
            }
            .gr-button:focus {
                border-color: rgb(147 197 253 / var(--tw-border-opacity));
                outline: none;
                box-shadow: var(--tw-ring-offset-shadow), var(--tw-ring-shadow), var(--tw-shadow, 0 0 #0000);
                --tw-border-opacity: 1;
                --tw-ring-offset-shadow: var(--tw-ring-inset) 0 0 0 var(--tw-ring-offset-width) var(--tw-ring-offset-color);
                --tw-ring-shadow: var(--tw-ring-inset) 0 0 0 calc(3px var(--tw-ring-offset-width)) var(--tw-ring-color);
                --tw-ring-color: rgb(191 219 254 / var(--tw-ring-opacity));
                --tw-ring-opacity: .5;
            }
            #advanced-btn {
                font-size: .7rem !important;
                line-height: 19px;
                margin-top: 12px;
                margin-bottom: 12px;
                padding: 2px 8px;
                border-radius: 14px !important;
            }
            #advanced-options {
                display: none;
                margin-bottom: 20px;
            }
            .footer {
                margin-bottom: 45px;
                margin-top: 35px;
                text-align: center;
                border-bottom: 1px solid #e5e5e5;
            }
            .footer>p {
                font-size: .8rem;
                display: inline-block;
                padding: 0 10px;
                transform: translateY(10px);
                background: white;
            }
            .dark .footer {
                border-color: #303030;
            }
            .dark .footer>p {
                background: #0b0f19;
            }
            .acknowledgments h4{
                margin: 1.25em 0 .25em 0;
                font-weight: bold;
                font-size: 115%;
            }
            .animate-spin {
                animation: spin 1s linear infinite;
            }
            @keyframes spin {
                from {
                    transform: rotate(0deg);
                }
                to {
                    transform: rotate(360deg);
                }
            }
            #share-btn-container {
                display: flex; padding-left: 0.5rem !important; padding-right: 0.5rem !important; background-color: #000000; justify-content: center; align-items: center; border-radius: 9999px !important; width: 13rem;
                margin-top: 10px;
                margin-left: auto;
            }
            #share-btn {
                all: initial; color: #ffffff;font-weight: 600; cursor:pointer; font-family: 'IBM Plex Sans', sans-serif; margin-left: 0.5rem !important; padding-top: 0.25rem !important; padding-bottom: 0.25rem !important;right:0;
            }
            #share-btn * {
                all: unset;
            }
            #share-btn-container div:nth-child(-n+2){
                width: auto !important;
                min-height: 0px !important;
            }
            #share-btn-container .wrap {
                display: none !important;
            }
            
            .gr-form{
                flex: 1 1 50%; border-top-right-radius: 0; border-bottom-right-radius: 0;
            }
            #prompt-container{
                gap: 0;
            }
            #prompt-text-input, #negative-prompt-text-input{padding: .45rem 0.625rem}
            #component-16{border-top-width: 1px!important;margin-top: 1em}
            .image_duplication{position: absolute; width: 100px; left: 50px}
    """
    
    block = gr.Blocks(css=css, title="HF Stable Diffusion gradio UI")
    
    examples = [
        [
            'A high tech solarpunk utopia in the Amazon rainforest',
            'low quality',
            9
        ],
        [
            'A pikachu fine dining with a view to the Eiffel Tower',
            'low quality',
            9
        ],
        [
            'A mecha robot in a favela in expressionist style',
            'low quality, 3d, photorealistic',
            9
        ],
        [
            'an insect robot preparing a delicious meal',
            'low quality, illustration',
            9
        ],
        [
            "A small cabin on top of a snowy mountain in the style of Disney, artstation",
            'low quality, ugly',
            9
        ],
    ]
    
    
    with block:
        gr.HTML(
            """
                <div style="text-align: center; margin: 0 auto;">
                  <div
                    style="
                      display: inline-flex;
                      align-items: center;
                      gap: 0.8rem;
                      font-size: 1.75rem;
                    "
                  >
                    <svg
                      width="0.65em"
                      height="0.65em"
                      viewBox="0 0 115 115"
                      fill="none"
                      xmlns="http://www.w3.org/2000/svg"
                    >
                      <rect width="23" height="23" fill="white"></rect>
                      <rect y="69" width="23" height="23" fill="white"></rect>
                      <rect x="23" width="23" height="23" fill="#AEAEAE"></rect>
                      <rect x="23" y="69" width="23" height="23" fill="#AEAEAE"></rect>
                      <rect x="46" width="23" height="23" fill="white"></rect>
                      <rect x="46" y="69" width="23" height="23" fill="white"></rect>
                      <rect x="69" width="23" height="23" fill="black"></rect>
                      <rect x="69" y="69" width="23" height="23" fill="black"></rect>
                      <rect x="92" width="23" height="23" fill="#D9D9D9"></rect>
                      <rect x="92" y="69" width="23" height="23" fill="#AEAEAE"></rect>
                      <rect x="115" y="46" width="23" height="23" fill="white"></rect>
                      <rect x="115" y="115" width="23" height="23" fill="white"></rect>
                      <rect x="115" y="69" width="23" height="23" fill="#D9D9D9"></rect>
                      <rect x="92" y="46" width="23" height="23" fill="#AEAEAE"></rect>
                      <rect x="92" y="115" width="23" height="23" fill="#AEAEAE"></rect>
                      <rect x="92" y="69" width="23" height="23" fill="white"></rect>
                      <rect x="69" y="46" width="23" height="23" fill="white"></rect>
                      <rect x="69" y="115" width="23" height="23" fill="white"></rect>
                      <rect x="69" y="69" width="23" height="23" fill="#D9D9D9"></rect>
                      <rect x="46" y="46" width="23" height="23" fill="black"></rect>
                      <rect x="46" y="115" width="23" height="23" fill="black"></rect>
                      <rect x="46" y="69" width="23" height="23" fill="black"></rect>
                      <rect x="23" y="46" width="23" height="23" fill="#D9D9D9"></rect>
                      <rect x="23" y="115" width="23" height="23" fill="#AEAEAE"></rect>
                      <rect x="23" y="69" width="23" height="23" fill="black"></rect>
                    </svg>
                    <h1 style="font-weight: 900; margin-bottom: 7px;margin-top:5px">
                      Stable Diffusion 2.1 Demo
                    </h1>
                  </div>
                  <p style="margin-bottom: 10px; font-size: 94%; line-height: 23px;">
                    Stable Diffusion 2.1 is the latest text-to-image model from StabilityAI. <a style="text-decoration: underline;" href="https://huggingface.co/spaces/stabilityai/stable-diffusion-1">Access Stable Diffusion 1 Space here</a><br>For faster generation and API
                    access you can try
                    <a
                      href="http://beta.dreamstudio.ai/"
                      style="text-decoration: underline;"
                      target="_blank"
                      >DreamStudio Beta</a
                    >.</a>
                  </p>
                </div>
            """
        )
        with gr.Group():
            with gr.Box():
                ## TODO: drop 'rounded', border, margin, and other options no longer supported (see log)
                with gr.Row(elem_id="prompt-container").style(mobile_collapse=False, equal_height=True):
                    with gr.Column():
                        prompt_control = gr.Textbox(
                            label="Enter your prompt",
                            show_label=False,
                            max_lines=1,
                            placeholder="Enter your prompt",
                            elem_id="prompt-text-input",
                        ).style(
                            border=(True, False, True, True),
                            rounded=(True, False, False, True),
                            container=False,
                        )
                        negative_control = gr.Textbox(
                            label="Enter your negative prompt",
                            show_label=False,
                            max_lines=1,
                            placeholder="Enter a negative prompt",
                            elem_id="negative-prompt-text-input",
                        ).style(
                            border=(True, False, True, True),
                            rounded=(True, False, False, True),
                            container=False,
                        )
                    btn = gr.Button("Generate image").style(
                        margin=False,
                        rounded=(False, True, True, False),
                        full_width=False,
                    )
    
            gallery = gr.Gallery(
                label="Generated images", show_label=False, elem_id="gallery"
            ).style(grid=[2], height="auto")
    
    
            with gr.Accordion("Advanced settings", open=False):
            #    gr.Markdown("Advanced settings are temporarily unavailable")
            #    samples = gr.Slider(label="Images", minimum=1, maximum=4, value=4, step=1)
            #    steps = gr.Slider(label="Steps", minimum=1, maximum=50, value=45, step=1)
                 guidance_control = gr.Slider(
                    label="Guidance Scale", minimum=0, maximum=31, value=GUIDANCE_SCALE, step=0.1
                 )
                 num_control = gr.Slider(
                    label="Number of images", minimum=1, maximum=10, value=2, step=1
                 )
                 
            #    seed = gr.Slider(
            #        label="Seed",
            #        minimum=0,
            #        maximum=2147483647,
            #        step=1,
            #        randomize=True,
            #    )

            input_controls = [prompt_control, negative_control, guidance_control, num_control]
            output_controls = [gallery]
            ex = gr.Examples(examples=examples, fn=infer,
                             inputs=input_controls,
                             outputs=output_controls, cache_examples=False)
            ex.dataset.headers = [""]
            negative_control.submit(infer, inputs=input_controls, outputs=output_controls, postprocess=False)
            prompt_control.submit(infer, inputs=input_controls, outputs=output_controls, postprocess=False)
            btn.click(infer, inputs=input_controls, outputs=output_controls, postprocess=False)
            
            #advanced_button.click(
            #    None,
            #    [],
            #    text,
            #    _js="""
            #    () => {
            #        const options = document.querySelector("body > gradio-app").querySelector("#advanced-options");
            #        options.style.display = ["none", ""].includes(options.style.display) ? "flex" : "none";
            #    }""",
            #)
            gr.HTML(
                """
                    <div class="footer">
                        <p>Model by <a href="https://huggingface.co/stabilityai" style="text-decoration: underline;" target="_blank">StabilityAI</a> - backend running JAX on TPUs due to generous support of <a href="https://sites.research.google/trc/about/" style="text-decoration: underline;" target="_blank">Google TRC program</a> - Gradio Demo by 🤗 Hugging Face
                        </p>
                    </div>
               """
            )
            with gr.Accordion(label="License", open=False):
                gr.HTML(
                    """<div class="acknowledgments">
                        <p><h4>LICENSE</h4>
    The model is licensed with a <a href="https://huggingface.co/stabilityai/stable-diffusion-2/blob/main/LICENSE-MODEL" style="text-decoration: underline;" target="_blank">CreativeML OpenRAIL++</a> license. The authors claim no rights on the outputs you generate, you are free to use them and are accountable for their use which must not go against the provisions set in this license. The license forbids you from sharing any content that violates any laws, produce any harm to a person, disseminate any personal information that would be meant for harm, spread misinformation and target vulnerable groups. For the full list of restrictions please <a href="https://huggingface.co/spaces/CompVis/stable-diffusion-license" target="_blank" style="text-decoration: underline;" target="_blank">read the license</a></p>
                        <p><h4>Biases and content acknowledgment</h4>
    Despite how impressive being able to turn text into image is, beware to the fact that this model may output content that reinforces or exacerbates societal biases, as well as realistic faces, pornography and violence. The model was trained on the <a href="https://laion.ai/blog/laion-5b/" style="text-decoration: underline;" target="_blank">LAION-5B dataset</a>, which scraped non-curated image-text-pairs from the internet (the exception being the removal of illegal content) and is meant for research purposes. You can read more in the <a href="https://huggingface.co/CompVis/stable-diffusion-v1-4" style="text-decoration: underline;" target="_blank">model card</a></p>
                   </div>
                    """
                )
                
    block.queue(concurrency_count=80, max_size=100).launch(max_threads=150)

                
#-------------------------------------------------------------------------------
# Runtime support

def main():
    """Entry point"""

    # Parse command line argument, show usage if --help given
    # TODO? auto_help=False
    main_app = Main(description=__doc__, skip_input=True,
                    boolean_options=[(BATCH_ARG, "Use batch mode--no UI"),
                                     (SERVER_ARG, "Run flask server"),
                                     (UI_ARG, "Show user interface")],
                    text_options=[(PROMPT_ARG, "Positive prompt"),
                                  (NEGATIVE_ARG, "Negative prompt")],
                    int_options=[(GUIDANCE_ARG, "Degree of fidelity to prompt (1-to-30 w/ 7 suggested)")])
    debug.trace_object(5, main_app)
    debug.assertion(main_app.parsed_args)
    #
    batch_mode = main_app.get_parsed_option(BATCH_ARG)
    server_mode = main_app.get_parsed_option(SERVER_ARG)
    ui_mode = main_app.get_parsed_option(UI_ARG, not (batch_mode or server_mode))
    prompt = main_app.get_parsed_option(PROMPT_ARG, PROMPT)
    negative_prompt = main_app.get_parsed_option(NEGATIVE_ARG, NEGATIVE_PROMPT)
    guidance = main_app.get_parsed_option(GUIDANCE_ARG, GUIDANCE_SCALE)
    # TODO2: BASENAME and NUM_IMAGES
    ## TODO: x_mode = main_app.get_parsed_option(X_ARG)
    debug.assertion(not (batch_mode and server_mode))

    # Invoke UI via HTTP unless in batch mode
    global sd_instance
    sd_instance = StableDiffusion(use_hf_api=server_mode)
    if batch_mode:
        images = infer(prompt, negative_prompt, guidance, skip_img_spec=True)
        for i, image_encoding in enumerate(images):
            # note: "encodes" UTF-8 text of base-64 encoding as bytes object for str, and then decodes into image bytes
            image_data = base64.decodebytes(image_encoding.encode())
            system.write_binary_file(f"{BASENAME}-{i + 1}.png", image_data)
        # TODO2: get list of files via infer()
        file_spec = " ".join(gh.get_matching_files(f"{BASENAME}*png"))  
        print(f"See {file_spec} for output image(s).")
    elif server_mode:
        debug.assertion(SD_URL)
        debug.assertion(not sd_instance.server_url)
        debug.trace_object(5, flask_app)
        flask_app.run(host=SD_URL, port=SD_PORT, debug=SD_DEBUG)
    else:
        debug.assertion(ui_mode)
        run_ui()


if __name__ == '__main__':
    debug.trace_current_context(level=TL.QUITE_DETAILED)
    main()



#!/usr/bin/env python3

"""A Corpus is a collection of Documents. Calculates IDF and full TF-IDF.

Vocabulary:
    stem - to stem (or "stemming") is to convert to a word from the original
        conjugated, plural, etc. to a simpler form.
            Examples:
                cats -> cats
                winningly -> win
    term - a one or multi word string, no stemming
    ngram - a one or multi word string that has been pre-processed and stemmed.
        Stemming may not be idempotent, so it's important to not re-process the
        ngram.
        bi-gram = two-word ngram
        tri-gram = three-word ngram
        n-gram = "n" word ngram
    keyword - contains the ngram, references to all the document(s) it is located
        in, and the locations within the document so that the original term can
        also be re-calculated.
    document - a body of text
    corpus - a collection of documents that are at least somewhat comparible.
        it is recommended to NOT have multiple languages within a single corpus.
    TF - Term Frequency within a document
    IDF - Inverse Document Frequency, the inverse frequency of the term across
        all documents in a corpus
    TF-IDF - a combination of the TF and IDF scores reflecting the "importance"
        of a term within a particular document
"""

## NOTE: See TPO for Thomas P. O'Hara's hacks
## TODO:
## - add missing docstrings for a few of the methods below
## - turn string keyword arguments from kw=default to
##     kw=None ... if kw is None: kw = default
## - Add debug.trace_fmt(level, ...) in place of unconditional sys.stderr.write's.
##

from __future__ import absolute_import, division

# Standard modules
import math
from collections import namedtuple
## TODO
## import os
import sys

# Installed modules
## TEST: from cachetools import LRUCache, cached
from functools import lru_cache
from mezcla import debug
from mezcla import system
from mezcla import misc_utils

## TODO
## # HACK: For relative imports to work in Python 3.6
## # See https://stackoverflow.com/questions/16981921/relative-imports-in-python-3.
## sys.path.append(os.path.dirname(os.path.abspath(__file__)))

## OLD
## # TPO: environment option for weight singleton occurrences low
## import os
## import sys
## PENALIZE_SINGLETONS = os.environ.get("PENALIZE_SINGLETONS", "False")
## gif str(PENALIZE_SINGLETONS).upper() in ["TRUE", "1"]:
##     sys.stderr.write("Penalizing singleton ngrams\n")
##     PENALIZE_SINGLETONS = True
## import sys

# Local modules

## OLD:
## from .document import Document
## from .preprocess import Preprocessor, clean_text
from mezcla.tfidf.document import Document, PENALIZE_SINGLETONS
from mezcla.tfidf.preprocess import Preprocessor, clean_text

CorpusKeyword = namedtuple('CorpusKeyword', ['term', 'ngram', 'score'])

SKIP_NORMALIZATION = system.getenv_bool("SKIP_NORMALIZATION", False,
                                        "Skip term/ngram normalization")
NGRAM_EPSILON = system.getenv_bool("NGRAM_EPSILON", 0.000001,
                                   "Occurrence count for unknown ngrams")

class Corpus(object):
    """A corpus is made up of Documents, and performs TF-IDF calculations on them.

    After initialization, add "documents" to the Corpus by adding text
    strings with a document "key". These will generate Document objects.

    Example:
        >>> from mezcla.tfidf.corpus import Corpus
        >>> c = Corpus(gramsize=2)
        >>> c['doc1'] = 'Mary had a little lamb.'
        >>> c['doc2'] = 'Hannible is not a lamb.'
        >>> c['doc3'] = 'The shark sleeps a little.'
        >>> c.df_freq('lamb')
        0
        >>> c.df_freq('little lamb')
        1
        >>> c.df_freq('a little')
        2
        >>> round(c.idf('a little'), 3)
        0.405
    """

    def __init__(self, min_ngram_size=None, max_ngram_size=None,
                 language=None, preprocessor=None,
                 gramsize=None, all_ngrams=None):
        """Initalize.

        Parameters:
            min_ngram_size (int-or-None):
                if integral, gives the lower bound for ngram sie
            max_ngram_size (int-or-None):
                if integral, gives the upper bound for ngram size
            language (str):
                Uses NLTK's stemmer and local stopwords files appropriately for the
                language.
                Check available languages with Preprocessor.supported_languages
            preprocessor (object):
                pass in a preprocessor object if you want to manually configure stemmer
                and stopwords
            all_ngrams (bool):
                if True, return all possible ngrams of size "gramsize" and smaller.
                else, only return keywords of exactly length "gramsize"
                Note: deprecated (use min_ngram_size instead).
            gramsize (int): number of words in a keyword
                deprecated: use max_ngram_size instead
        """
        debug.assertion(not (gramsize and max_ngram_size))
        debug.assertion(not (all_ngrams and min_ngram_size))
        self.__documents = {}
        self.__document_occurrences = {}
        ## OLD: self.__gramsize = gramsize
        self.__gramsize = (max_ngram_size or gramsize)
        self.__max_raw_frequency = None
        self.__max_rel_doc_frequency = None
        self.__max_doc_frequency = None
        if preprocessor:
            self.preprocessor = preprocessor
        else:
            self.preprocessor = Preprocessor(
                language=language, min_ngram_size=min_ngram_size, max_ngram_size=max_ngram_size,
                ## TODO: remove deprecated
                gramsize=gramsize, all_ngrams=all_ngrams)

    def __contains__(self, document_id):
        """A Corpus contains Document ids."""
        return document_id in self.__documents

    def __len__(self):
        """Length of a Corpus is the number of Documents it holds."""
        return len(self.__documents)

    def __getitem__(self, document_id):
        """Fetch a Document from the Corpus by its id."""
        return self.__documents[document_id]

    def __setitem__(self, document_id, text):
        """Add a Document to the Corpus using a unique id key."""
        text = clean_text(text)
        self.__documents[document_id] = Document(text, self.preprocessor)

    @property
    def gramsize(self):
        """Number of words in the ngram. Not editable post init."""
        return self.__gramsize

    def keys(self):
        """The document ids in the corpus."""
        return self.__documents.keys()

    @property
    def max_raw_frequency(self):
        """Highest frequency across all Documents in the Corpus."""
        ## TPO: caches value
        ## OLD: return max([_.max_raw_frequency for _ in self.__documents.values()])
        if self.__max_raw_frequency is None:
            self.__max_raw_frequency = max([_.max_raw_frequency for _ in self.__documents.values()])
        return self.__max_raw_frequency

    @lru_cache()
    def count_doc_occurrences(self, ngram):
        """Count the number of documents the corpus has with the matching ngram."""
        # Note: caches the values as called multiple times for TFIDF calculation
        if ngram in self.__document_occurrences:
            count = self.__document_occurrences[ngram]
        else:
            ## OLD: count = sum([1 if ngram in doc else 0 for doc in self.__documents.values()])
            count = sum([(1 + NGRAM_EPSILON) if ngram in doc else 0 for doc in self.__documents.values()])
        if ((count == 1) and PENALIZE_SINGLETONS):
            count = 0
        if (count == 0):
            count = NGRAM_EPSILON
        return count

    ## TPO
    @property
    def max_rel_doc_frequency(self):
        """"Highest relative document frequency for all ngrams in the corpus"""
        if self.__max_rel_doc_frequency is None:
            ## BAD: max_doc_frequency = max([self.count_doc_occurrences(ngram) for ngram in self.__documents.values()])
            max_doc_frequency = max([self.count_doc_occurrences(ngram) for doc in self.__documents.values() for ngram in doc.keywordset])
            self.__max_rel_doc_frequency = max_doc_frequency / float(len(self))
        return self.__max_rel_doc_frequency
    
    ## TPO
    @property
    def max_doc_frequency(self):
        """"Highest relative document frequency for all ngrams in the corpus"""
        if self.__max_doc_frequency is None:
            max_doc_frequency = max([self.count_doc_occurrences(ngram) for doc in self.__documents.values() for ngram in doc.keywordset])
            self.__max_doc_frequency = max_doc_frequency
        return self.__max_doc_frequency
    
    def df_freq(self, ngram):
        """Return document frequency (DF) count for NGRAM"""
        return self.count_doc_occurrences(ngram)

    ## OLD: def df_norm(self, ngram, normalize_term=True):
    def df_norm(self, ngram, normalize_term=None):
        """Return DF in range [epsilon, 1] for NGRAM, with NORMALIZE_TERM by default.
        Note: The denominator is the max doc frequency rather than number of documents,
        so the values of df_norm are better comparable for different collections."""
        if normalize_term is None:
            normalize_term = not SKIP_NORMALIZATION
        if normalize_term:
            ngram = self.preprocessor.normalize_term(ngram)
        rel_doc_freq = 0
        if self.max_rel_doc_frequency > 0:
            ## BAD: rel_doc_freq = (self.df_freq(ngram) / self.max_rel_doc_frequency)
            rel_doc_freq = (self.df_freq(ngram) / self.max_doc_frequency)
        return rel_doc_freq

    def idf_basic(self, ngram):
        """Returns IDF based on log of relative document frequency"""
        debug.assertion(self.count_doc_occurrences(ngram) >= 1)
        if self.count_doc_occurrences(ngram) == 0:
            raise Exception(ngram)
        num_occurrences = self.count_doc_occurrences(ngram)
        ## TPO: TEST
        ## # HACK: give singletons a max DF to lower IDF score
        ## if (num_occurrences == 1) and PENALIZE_SINGLETONS:
        ##     num_occurrences = len(self.__documents)
        idf = math.log(float(len(self)) / num_occurrences)
        debug.trace_fmt(8, "idf_basic({ng} len(self)={l} max_doc_occ={mdo} num_occ={no} idf={idf})\n",
                        ng=ngram, l=len(self), mdo=self.max_doc_frequency, no=num_occurrences, idf={idf})
        return idf

    def idf_freq(self, ngram):
        """Returns inverse proper of DF (not N/DF)"""
        debug.assertion(self.count_doc_occurrences(ngram) >= 1)
        if self.count_doc_occurrences(ngram) == 0:
            raise Exception(ngram)
        num_occurrences = self.count_doc_occurrences(ngram)
        idf = 1 / num_occurrences
        debug.trace_fmt(8, "idf_freq({ng} len(self)={l} max_doc_occ={mdo} num_occ={no} idf={idf})\n",
                        ng=ngram, l=len(self), mdo=self.max_doc_frequency, no=num_occurrences, idf={idf})
        return idf

    def idf_smooth(self, ngram):
        """Returns IDF using simple smoothing with add-1 relative frequency (prior to log)"""
        debug.assertion(self.count_doc_occurrences(ngram) >= 1)
        idf = math.log(1 + (float(len(self)) / self.count_doc_occurrences(ngram)))
        debug.trace_fmt(8, "idf_smooth({ng} len(self)={l} doc_occ={do} idf={idf})\n",
                        ng=ngram, l=len(self), do=self.count_doc_occurrences(ngram), idf={idf})
        return idf

    def idf_max(self, ngram):
        ## TODO: make sure this interpretation makes sense; also, make sure float conversion not required
        """Use maximum ngram TF in place of N and also perform add-1 smoothing"""
        debug.assertion(self.count_doc_occurrences(ngram) >= 1)
        idf = math.log(1 + self.max_raw_frequency / self.count_doc_occurrences(ngram))
        debug.trace_fmt(8, "idf_smooth({ng} len(self)={l} doc_occ={do} idf={idf})\n",
                        ng=ngram, l=len(self), do=self.count_doc_occurrences(ngram), idf={idf})
        return idf

    def idf_probabilistic(self, ngram):
        """Returns IDF via probabilistic interpretation: log((N - d)/d), where d is the document occcurrence count for the NGRAM"""
        debug.assertion(self.count_doc_occurrences(ngram) >= 1)
        ## TODO: shouldn't this be (float(len(self) / num_doc_occurrences))
        num_doc_occurrences = self.count_doc_occurrences(ngram)
        idf = math.log(float(len(self) - num_doc_occurrences) / num_doc_occurrences)
        debug.trace_fmt(8, "idf_smooth({ng} len(self)={l} doc_occ={do} idf={idf})\n",
                        ng=ngram, l=len(self), do=num_doc_occurrences, idf={idf})
        return idf

    def idf(self, ngram, idf_weight='basic'):
        """Inverse document frequency (IDF) indicates ngram common-ness across the Corpus."""
        ## OLD: return w/ elif
        ## if idf_weight == 'smooth':
        ##     return self.idf_smooth(ngram)
        ## elif idf_weight == 'basic':
        ##     return self.idf_basic(ngram)
        ## elif idf_weight == 'max':
        ##     return self.idf_max(ngram)
        ## elif idf_weight == 'prob':
        ##     return self.idf_probabilistic(ngram)
        if idf_weight == 'freq':
            return self.idf_freq(ngram)
        if idf_weight == 'smooth':
            return self.idf_smooth(ngram)
        if idf_weight == 'basic':
            return self.idf_basic(ngram)
        if idf_weight == 'max':
            return self.idf_max(ngram)
        if idf_weight == 'prob':
            return self.idf_probabilistic(ngram)
        ## TODO: allow for raw frequency
        if idf_weight == 'freq':
            return self.idf_freq(ngram)
        raise ValueError("Invalid idf_weight: " + idf_weight)

    def tf_idf(self, term, document_id=None, text=None, idf_weight='basic', tf_weight='basic',
               ## OLD: normalize_term=True):
               normalize_term=None):
        """TF-IDF score. Must specify a document id (within corpus) or pass text body."""
        assert document_id or text
        if normalize_term is None:
            normalize_term = not SKIP_NORMALIZATION
        document = None
        if document_id:
            document = self[document_id]
        if text:
            text = clean_text(text)
            document = Document(text, self.preprocessor)
        if normalize_term:
            ngram = self.preprocessor.normalize_term(term)
        else:
            ngram = term
        score = document.tf(ngram, tf_weight=tf_weight) * self.idf(ngram, idf_weight=idf_weight)
        return CorpusKeyword(document[ngram], ngram, score)

    @lru_cache()
    def get_keywords(self, document_id=None, text=None, idf_weight='basic',
                     tf_weight='basic', limit=100):
        """Return a list of keywords with TF-IDF scores. Defaults to the top 100."""
        assert document_id or text
        document = None
        if document_id:
            document = self[document_id]
        if text:
            text = clean_text(text)
            document = Document(text, self.preprocessor)
        out = []
        for ngram, kw in document.keywordset.items():
            score = document.tf(ngram, tf_weight=tf_weight) * \
                self.idf(ngram, idf_weight=idf_weight)
            out.append(CorpusKeyword(kw, ngram, score))
        out.sort(key=lambda x: x.score, reverse=True)
        return out[:limit]


def new_main():
    """New entry point for script: counts ngrams in input file (one line per document)"""
    # TODO: reconcile with code in ngrm_tfidf.main; parameterize
    if (sys.argv[1] == "--help"):
        system.print_error("Usage: {prog} [--help] corpus-file".format(prog=sys.argv[0]))
        return
    input_file = sys.argv[1]

    # Create corpus from line documents in file
    corp = Corpus(min_ngram_size=2, max_ngram_size=4)
    doc_IDs = []
    for i, doc_text in enumerate(system.read_lines(input_file)):
        doc_IDs.append(f"line-doc{i + 1}")
        corp[doc_IDs[i]] = doc_text
               
    # Output top ngrams for each document
    SAMPLE_SIZE = system.getenv_int("SAMPLE_SIZE", 10,
                                    "Number of ngram samples to show")
    print(f"top {SAMPLE_SIZE} ngrams in each document")
    for i in range(len(doc_IDs)):
        tuples = corp.get_keywords(document_id=doc_IDs[i], limit=SAMPLE_SIZE)
        # note: ignores empty tokens
        ngram_info = [(ng.ngram + ":" + system.round_as_str(ng.score, 3)) for ng in tuples if ng.ngram.strip()]
        ngram_spec = "{" + ",".join(ngram_info) + "}"
        print(f"{doc_IDs[i]}:\t{ngram_spec}")
    return


def main():
    """Entry point for script: just runs a simple test"""
    # TODO: rename main as simple_test?
    if (len(sys.argv) > 1):
        new_main()
        return

    # Create simple corpus with two documents
    c = Corpus(min_ngram_size=2, max_ngram_size=2)
    doc_text = ["abc  def  ghi  jkl  mno         ",
                "abc  def       jkl  mno         ",
                "abc  def                pdq  rst"]
    ngrams = set()
    for i in range(len(doc_text)):
        doc_id = ("doc" + str(i + 1))
        c[doc_id] = doc_text[i]
        debug.trace_fmt(1, "{id}:\t{doc_text}", id=doc_id, doc_text=doc_text[i])
        words = doc_text[i].split()
        debug.assertion(len(words) > 1)
        ngrams.update({(words[i] + " " + words[i + 1]) for i in range(len(words) - 1)})
    all_ngrams = sorted(ngrams)

    # Show combined set of ngrams
    debug.trace_fmt(1, "ngrams:\t{spec}", spec="\t".join(ng.replace(" ", "_") for ng in sorted(all_ngrams)))

    # Show per-document TF for each ngram
    for docid in sorted(c.keys()):
        debug.trace_fmt(1, "{id}TF:\t{spec}", id=docid,
                        spec="\t".join([str(c[docid].tf_freq(ng)) for ng in all_ngrams]))

    # Show IDF, and TF-IDF for each ngram
    debug.trace_fmt(1, "IDF:   \t{spec}", spec="\t".join([system.round_as_str(c.idf(ng), 3) for ng in all_ngrams]))
    debug.assertion(misc_utils.is_close(c.idf("abc def"), 0))
    debug.assertion(c.idf("jkl mno") < c.idf("pdq rst"))
    debug.assertion(c.idf("def ghi") == c.idf("ghi jkl") == c.idf("pdq rst"))
        
#-------------------------------------------------------------------------------
    
if __name__ == '__main__':
    system.print_stderr(f"Warning: {__file__} is not intended to be run standalone. A simple test will be run.")
    main()



#! /usr/bin/env python
# 
# Text categorized based on BERT model. This allows for multiclass categorization. The
# input is a tabbed delimited file of the format:
#     category<TAB>token1<DELIM>token2<DELIM>...tokenN
# where DELIM is either whitespace and/or punctuation (e.g., ' ' or '. ' or ' $').
# The tokens are alphanumeric as well as including dashes and underscores.
#
# Notes:
# - This is based on following online article:
#       https://analyticsindiamag.com/step-by-step-guide-to-implement-multi-class-classification-with-bert-tensorflow
# - Godawful code organization (e.g., doubly embedded functions).
# - Default text categories:
#      Politics: 0
#      Technology: 1
#      Entertainment: 2
#      Business: 3
#
# - Data Preprocessing
#
#   The BERT model accepts only a specific type of input and the datasets are usually structured to have have the following four features:
#      guid : A unique id that represents an observation.
#      text_a : The text we need to classify into given categories
#      text_b: It is used when we're training a model to understand the relationship between sentences and it does not apply for classification problems.
#      label: It consists of the labels or classes or categories that a given text belongs to.
#
#  In our dataset we have text_a and label. The code will create objects for each of the above mentioned features for all the records in our dataset using the InputExample class provided in the BERT library.
#
# - To cache the tensorflow hub version of the BERT data, use the following steps,
#   based on https://medium.com/@xianbao.qian/how-to-run-tf-hub-locally-without-internet-connection-4506b850a915:
#       import hashlib
#       from mezcla import glue_helpers as gh
#       model = "uncased_L-12_H-768_A-12"
#       # -or-: model = "uncased_L-8_H-512_A-8"
#       BERT_MODEL_HUB = f"https://tfhub.dev/google/{model}/1"
#       # take 2:
#       #    model = "bert_uncased_L-8_H-512_A-8"
#       #    BERT_MODEL_HUB = f"https://tfhub.dev/google/small_bert/{model}/2
#       temp_dir = "/tmp"
#       # -or-: temp_dir = f"{os.environ.get('HOME', '.')}/temp"
#       local_hub_dir = f"{temp_dir}/tf-hub"
#       gh.issue(f"mkdir {local_hub_dir}")
#       os.environ["TFHUB_CACHE_DIR"] = local_hub_dir
#       cache_subdir = hashlib.sha1(BERT_MODEL_HUB.encode("utf8")).hexdigest()
#       gh.issue(f"mkdir {local_hub_dir}/{cache_subdir}")
#       tf_hub_url = (BERT_MODEL_HUB.                                                        \
#                         replace("https://tfhub.dev/google/",                               \
#                                 "https://storage.googleapis.com/bert_models/2020_02_20/"). \
#                         replace("/1", ".zip").replace("/2", ".zip").                       \
#                         replace("/google", "").replace("/small_bert", "))
#       temp_archive = f"{temp_dir}/{model}.zip"
#       gh.issue(f"curl {tf_hub_url} > {temp_archive}")
#       gh.issue(f"cd {local_hub_dir}/{cache_subdir}; unzip {temp_archive}")
#       print([BERT_MODEL_HUB, local_hub_dir, cache_subdir, temp_archive])
#   =>
#   ['https://tfhub.dev/google/uncased_L-12_H-768_A-12/1', '/tmp/tf-hub', '627e8a1af062e84ad50e5e82c59b0f86c3a5f3df', '/tmp/uncased_L-12_H-768_A-12.zip']
#   TODO: convert replace sequence to re.sub
#
#  For hub instance "https://tfhub.dev/google/uncased_L-12_H-768_A-12/1",
#  this downloads https://storage.googleapis.com/tfhub-modules/2020_02_20/uncased_L-12_H-768_A-12.zip"
#  and places the zipped contents in /tmp/tf-hub.
#
#--------------------------------------------------------------------------------
# TODO:
# - Determine the depencies between the versions of TensorFlow and BERT. This
#   doesn't work with TensorFlow 2.0 and BERT from 2018.
#

"""Run text categorization over input file (or Predict the News Category Hackathon)"""

# Standard packages
from datetime import datetime
## OLD: import hashlib
## OLD: import os

# Installed packages
import pandas as pd
import tensorflow as tf
import tensorflow_hub as hub
from sklearn.model_selection import train_test_split
# Import BERT modules
# note: need to use bert-tenorflow version 1.0.1
import bert
from bert import run_classifier

# Local packages
from mezcla import debug
from mezcla.main import Main
from mezcla import system
from mezcla import glue_helpers as gh

# Show TensorFlow info
# note: 1.15.0 and 0.7.0 for blog example
print("tensorflow version : ", tf.__version__)                # pylint: disable=no-member
print("tensorflow_hub version : ", hub.__version__)

# Set the output directory for saving model file
USER = system.getenv_text("USER", "user")
BASE_DIR = system.getenv_text("BASE_DIR", ".")
TRAINING_FILE = "Data_Train.xls"
TESTING_FILE = "Data_Test.xls"
SUBMISSION_FILE = "submission_bert.xls"
TYPICAL_INPUT_DIR = system.form_path("/home", USER, "data", "hackathon-news-prediction")
DEFAULT_INPUT_DIR = "." if system.file_exists(TRAINING_FILE) else TYPICAL_INPUT_DIR
INPUT_DIR = system.getenv_text("INPUT_DIR", DEFAULT_INPUT_DIR)
TRAINING_PATH = system.form_path(INPUT_DIR, TRAINING_FILE)
TESTING_PATH = system.form_path(INPUT_DIR, TESTING_FILE)
OUTPUT_DIR = system.getenv_text("OUTPUT_DIR",
                                system.form_path(INPUT_DIR, "output"))
SHOW_PLOTS = system.getenv_bool("SHOW_PLOTS", False)
TFHUB_CACHE_DIR = system.getenv_text("TFHUB_CACHE_DIR", "/tmp/tf-hub",
                                     "Directory for cache of models from Tensorflow Hub")
if TFHUB_CACHE_DIR:
    debug.trace_fmt(3, "Using local TensorFlow Hub cache: {c}", c=TFHUB_CACHE_DIR)
BERT_MODEL_HUB = system.getenv_text("BERT_MODEL_HUB", "https://tfhub.dev/google/bert_uncased_L-12_H-768_A-12/1",
                                    "Model to use from Tensorflow Hub")
USER_HOME = system.getenv_text("HOME", ".",
                               "User home directory")
## TODO
## USER = system.getenv_text("USER", "joe",
##                           "User ID")

# Whether or not to clear/delete the directory and create a new one
DO_DELETE = system.getenv_bool("DO_DELETE", False)

# Do sanity check on input files
debug.assertion(system.non_empty_file(TRAINING_PATH))
debug.assertion(system.non_empty_file(TESTING_PATH))

# Delete previous run if desired. The (re-)create the output directory if needed.
debug.assertion(system.is_directory(INPUT_DIR))
if DO_DELETE:
    try:
        gh.issue("/bin/rm --verbose --recursive {od}", od=OUTPUT_DIR)
    except:
        system.print_stderr("Problem deleting dir: {od}", od=OUTPUT_DIR)
if not system.is_directory(OUTPUT_DIR):
    system.create_directory(OUTPUT_DIR)
print('***** Model output directory: {} *****'.format(OUTPUT_DIR))
    

class Script(Main):
    """Adhoc script class (e.g., no I/O loop, just run calls)"""
    # This is a path to an uncased (all lowercase) version of BERT
    MAX_SEQ_LENGTH = 128         # Sequences have at most 128 tokens.
    # These hyperparameters are copied from this colab notebook (https://colab.sandbox.google.com/github/tensorflow/tpu/blob/master/tools/colab/bert_finetuning_with_cloud_tpus.ipynb)
    BATCH_SIZE = 32
    LEARNING_RATE = 2e-5
    NUM_TRAIN_EPOCHS = 3.0
    # Warmup is a period of time where the learning rate is small and gradually increases--usually helps training.
    WARMUP_PROPORTION = 0.1
    # Model configs
    SAVE_CHECKPOINTS_STEPS = 1000
    SAVE_SUMMARY_STEPS = 100
    #
    DATA_COLUMN = 'STORY'
    LABEL_COLUMN = 'SECTION'
    # The list containing all the classes (train['SECTION'].unique())
    LABEL_LIST = [0, 1, 2, 3]
    #
    tokenizer = None
    estimator = None

    def setup(self):
        """Check results of command line processing"""
        debug.trace_fmtd(5, "Script.setup(): self={s}", s=self)
        # Make CUDA settings (n.b., very idiosyncratic)
        # Notes:
        # 1. The actual library file are based on the GPU: see NVIDIA SDK.
        # 2. Unfortunately, BERT requires Tensorflow 1.15 which uses old CUDA.
        ## BAD: pkg_dir = gh.form_path(HOME, USER, ".conda", "pkgs")
        pkg_dir = gh.form_path(USER_HOME, ".conda", "pkgs")
        system.setenv("LD_LIBRARY_PATH", f"/usr/lib/x86_64-linux-gnu:/usr/lib/nvidia-450:/usr/local/cuda-11.1/lib64:{pkg_dir}/cudatoolkit-10.1.243-h6bb024c_0/lib:{pkg_dir}/cudnn-7.6.5-cuda10.1_0/lib:/usr/local/misc/lib/cuda-repo-ubuntu1604-10-0-local-10.0.130-410.48_1.0-1:/usr/local/cuda/extras/CUPTI/lib64")
        # Show environment and this object instance
        debug.trace(4, gh.run("printenv"))
        debug.trace_object(5, self, label="Script instance")

    def create_tokenizer_from_hub_module(self):
        """Get the vocab file and casing info from the Hub module."""
        debug.trace(5, "create_tokenizer_from_hub_module()")
        debug.reference_var(self)
        with tf.Graph().as_default():
            bert_module = hub.Module(BERT_MODEL_HUB)
            ## TODO
            ## try:
                # Take 1: Get from Tensorflow Hub server
                ## bert_module = hub.Module(BERT_MODEL_HUB)
            ## except:
                ## raise NotImplementedError()
                ## # Take 2: Get from cache (e.g., /home/tomohara/temp/tf-hub/3843e19587d721e7cfc9d9cfe564ca3da192ffc9)
                ## # TODO: get the module conversion working
                ## cache_subdir = hashlib.sha1(BERT_MODEL_HUB.encode("utf8")).hexdigest()
                ## cache_path = os.path.join(TFHUB_CACHE_DIR, cache_subdir)
                ## bert_module = hub.saved_model_module.create_module_spec_from_saved_model(cache_path)
            tokenization_info = bert_module(signature="tokenization_info", as_dict=True)
            with tf.Session() as sess:
                vocab_file, do_lower_case = sess.run(
                    [tokenization_info["vocab_file"],
                     tokenization_info["do_lower_case"]])
                debug.trace_expr(5, bert_module, vocab_file, do_lower_case)
        tokenizer = bert.tokenization.FullTokenizer(
            vocab_file=vocab_file, do_lower_case=do_lower_case)
        debug.trace(4, f"create_tokenizer_from_hub_module() => {tokenizer}")
        return tokenizer
            

    def create_model(self, is_predicting, input_ids, input_mask, segment_ids,
                     labels, num_labels):
        """Load the BERT model for fine-tuning."""
        debug.trace(5, f"create_model({tuple([is_predicting, input_ids, input_mask, segment_ids, labels, num_labels])})")
        debug.reference_var(self)
        bert_module = hub.Module(
            BERT_MODEL_HUB,
            trainable=True)
        bert_inputs = dict(
            input_ids=input_ids,
            input_mask=input_mask,
            segment_ids=segment_ids)
        bert_outputs = bert_module(
            inputs=bert_inputs,
            signature="tokens",
            as_dict=True)
    
        # Use "pooled_output" for classification tasks on an entire sentence.
        # Use "sequence_outputs" for token-level output.
        output_layer = bert_outputs["pooled_output"]
        hidden_size = output_layer.shape[-1].value
      
        # Create our own layer to tune for politeness data.
        # TODO: Make this optional (probably specific to Hackathon corpus genre)
        output_weights = tf.get_variable(
            "output_weights", [num_labels, hidden_size],
            initializer=tf.truncated_normal_initializer(stddev=0.02))
        output_bias = tf.get_variable(
            "output_bias", [num_labels], initializer=tf.zeros_initializer())
      
        with tf.variable_scope("loss"):
      
            # Dropout helps prevent overfitting
            output_layer = tf.nn.dropout(output_layer, keep_prob=0.9)
      
            logits = tf.matmul(output_layer, output_weights, transpose_b=True)
            logits = tf.nn.bias_add(logits, output_bias)
            log_probs = tf.nn.log_softmax(logits, axis=-1)
      
            # Convert labels into one-hot encoding
            one_hot_labels = tf.one_hot(labels, depth=num_labels, dtype=tf.float32)
      
            predicted_labels = tf.squeeze(tf.argmax(log_probs, axis=-1, output_type=tf.int32))
  
            # If predicting, return predicted labels and the probabiltiies.
            if is_predicting:
                model = (predicted_labels, log_probs)
                debug.trace(5, f"create_model() => {model}; first return")
                return model
      
            # If in train/eval, compute loss between predicted and actual label
            # Note: returns loss, predicted labels, and the log of the label probs.
            per_example_loss = -tf.reduce_sum(one_hot_labels * log_probs, axis=-1)
            loss = tf.reduce_mean(per_example_loss)
            model = (loss, predicted_labels, log_probs)
            debug.trace(5, f"create_model() => {model}; second return")
            return model

    def model_fn_builder(self, num_labels, learning_rate, num_train_steps,
                         num_warmup_steps):
        """Returns `model_fn` closure for TPUEstimator."""
        # model_fn_builder actually creates our model function
        # using the passed parameters for num_labels, learning_rate, etc.
        debug.trace(5, f"model_fn_builder({tuple([num_labels, learning_rate, num_train_steps, num_warmup_steps])})")

        def model_fn(features, labels, mode, params):  # pylint: disable=unused-argument
            """The `model_fn` for TPUEstimator."""
            debug.trace(5, f"model_fn({tuple([features, labels, mode, params])})")
            input_ids = features["input_ids"]
            input_mask = features["input_mask"]
            segment_ids = features["segment_ids"]
            label_ids = features["label_ids"]
            is_predicting = (mode == tf.estimator.ModeKeys.PREDICT)
    
            # TRAIN and EVAL
            if not is_predicting:
                (loss, predicted_labels, log_probs) = self.create_model(   # pylint: disable=unbalanced-tuple-unpacking
                    is_predicting, input_ids, input_mask, segment_ids, label_ids, num_labels)
    
                train_op = bert.optimization.create_optimizer(
                    loss, learning_rate, num_train_steps, num_warmup_steps, use_tpu=False)
    
                # Calculate evaluation metrics. 
                def metric_fn(label_ids, predicted_labels):
                    """Evaluate LABEL_IDS against PREDICTED_LABELS"""
                    debug.trace(5, f"metric_fn({tuple([label_ids, predicted_labels])})")
                    accuracy = tf.metrics.accuracy(label_ids, predicted_labels)
                    true_pos = tf.metrics.true_positives(label_ids,
                                                         predicted_labels)
                    true_neg = tf.metrics.true_negatives(label_ids,
                                                         predicted_labels)   
                    false_pos = tf.metrics.false_positives(label_ids,
                                                           predicted_labels)  
                    false_neg = tf.metrics.false_negatives(label_ids,
                                                           predicted_labels)
              
                    return {
                        "eval_accuracy": accuracy,
                        "true_positives": true_pos,
                        "true_negatives": true_neg,
                        "false_positives": false_pos,
                        "false_negatives": false_neg
                    }
      
                eval_metrics = metric_fn(label_ids, predicted_labels)
      
                if mode == tf.estimator.ModeKeys.TRAIN:
                    return tf.estimator.EstimatorSpec(mode=mode,
                                                      loss=loss,
                                                      train_op=train_op)
                return tf.estimator.EstimatorSpec(mode=mode,
                                                  loss=loss,
                                                  eval_metric_ops=eval_metrics)
            (predicted_labels, log_probs) = self.create_model(
                is_predicting, input_ids, input_mask, segment_ids, label_ids, num_labels)
    
            predictions = {
                'probabilities': log_probs,
                'labels': predicted_labels
            }
            return tf.estimator.EstimatorSpec(mode, predictions=predictions)

        # Return the actual model function in the closure
        return model_fn

    def getPrediction(self, in_sentences):
        """A method to get predictions"""
        debug.trace(5, f"getPrediction({in_sentences})")

        # A list to map the actual labels to the predictions
        labels = ["Politics", "Technology", "Entertainment", "Business"]
      
        # Transforming the test data into BERT accepted form
        input_examples = [run_classifier.InputExample(guid="", text_a=x, text_b=None, label=0) for x in in_sentences] 
        
        # Creating input features for Test data
        input_features = run_classifier.convert_examples_to_features(input_examples, self.LABEL_LIST, self.MAX_SEQ_LENGTH, self.tokenizer)
      
        # Predicting the classes 
        predict_input_fn = run_classifier.input_fn_builder(features=input_features, seq_length=self.MAX_SEQ_LENGTH, is_training=False, drop_remainder=False)
        predictions = self.estimator.predict(predict_input_fn)
        return [(sentence, prediction['probabilities'], prediction['labels'], labels[prediction['labels']]) for sentence, prediction in zip(in_sentences, predictions)]
      
      
    def run_main_step(self):
        """Main processing step"""
        debug.trace_fmtd(5, "Script.run_main_step(): self={s}", s=self)

        # Load the data and also split the training set in to training and validation sets.
        train = pd.read_excel(system.form_path(INPUT_DIR, TRAINING_PATH))
        test = pd.read_excel(system.form_path(INPUT_DIR, TESTING_PATH))
        train, val =  train_test_split(train, test_size=0.2, random_state=100)

        # Show training set sample
        debug.trace_fmt(4, "Train sample:\n{s}", s=train.head(5))
        debug.trace_fmt(4, "Test sample:\n{s}", s=test.head(5))
        print("Training Set Shape :", train.shape)
        print("Validation Set Shape :", val.shape)
        print("Test Set Shape :", test.shape)

        # Features in the dataset
        # ex: Index(['STORY', 'SECTION'], dtype='object')
        debug.trace_fmt(4, "Train columns: {c}", c=train.columns)

        # unique classes
        # ex; array([3, 1, 2, 0])
        debug.trace_fmt(4, "Unique training classes: {c}",
                        c=train['SECTION'].unique())
        
        # Distribution of classes
        # ex: <matplotlib.axes._subplots.AxesSubplot at 0x7ff5921f9ba8>
        if SHOW_PLOTS:
            train['SECTION'].value_counts().plot(kind='bar')

        # Compute train and warmup steps from batch size
        ## BAD: num_train_steps = int(len(train_features) / self.BATCH_SIZE * self.NUM_TRAIN_EPOCHS)
        num_train_steps = int(len(train) / self.BATCH_SIZE * self.NUM_TRAIN_EPOCHS)
        num_warmup_steps = int(num_train_steps * self.WARMUP_PROPORTION)
        
        # Specify output directory and number of checkpoint steps to save
        run_config = tf.estimator.RunConfig(
            model_dir=OUTPUT_DIR,
            save_summary_steps=self.SAVE_SUMMARY_STEPS,
            save_checkpoints_steps=self.SAVE_CHECKPOINTS_STEPS)
        
        # Specify output directory and number of checkpoint steps to save
        run_config = tf.estimator.RunConfig(
            model_dir=OUTPUT_DIR,
            save_summary_steps=self.SAVE_SUMMARY_STEPS,
            save_checkpoints_steps=self.SAVE_CHECKPOINTS_STEPS)

        # Initializing the model and the estimator
        model_fn = self.model_fn_builder(
            num_labels=len(self.LABEL_LIST),
            learning_rate=self.LEARNING_RATE,
            num_train_steps=num_train_steps,
            num_warmup_steps=num_warmup_steps)
        #
        self.estimator = tf.estimator.Estimator(
            model_fn=model_fn,
            config=run_config,
            params={"batch_size": self.BATCH_SIZE})
        
        # Read in training data
        # See 'Data preprocessing' notes in the header comments
        train_InputExamples = train.apply(lambda x:
                                          bert.run_classifier.InputExample(
                                              guid=None,
                                              text_a=x[self.DATA_COLUMN], 
                                              text_b=None, 
                                              label=x[self.LABEL_COLUMN]),
                                          axis=1)
        val_InputExamples = val.apply(lambda x:
                                      bert.run_classifier.InputExample(
                                          guid=None, 
                                          text_a=x[self.DATA_COLUMN], 
                                          text_b=None, 
                                          label=x[self.LABEL_COLUMN]),
                                      axis=1)
        debug.trace_object(5, train_InputExamples, "train_InputExamples")
        debug.trace_fmt(4, "Row 0 - guid of training set: {r}", r=train_InputExamples.iloc[0].guid)
        debug.trace_fmt(4, "__________\nRow 0 - text_a of training set {t}: ", t=train_InputExamples.iloc[0].text_a)
        debug.trace_fmt(4, "__________\nRow 0 - text_b of training set: {t}", t=train_InputExamples.iloc[0].text_b)
        debug.trace_fmt(4, "__________\nRow 0 - label of training set: {l}", l=train_InputExamples.iloc[0].label)

        # Convert to BERT format
        self.tokenizer = self.create_tokenizer_from_hub_module()
        debug.trace_fmt(4, "row 0 tokenized: {t}", t=self.tokenizer.tokenize(train_InputExamples.iloc[0].text_a))

        # Convert train and validation features to InputFeatures that BERT understands.
        train_features = bert.run_classifier.convert_examples_to_features(
            train_InputExamples, self.LABEL_LIST, self.MAX_SEQ_LENGTH, self.tokenizer)

        val_features = bert.run_classifier.convert_examples_to_features(
            val_InputExamples, self.LABEL_LIST, self.MAX_SEQ_LENGTH, self.tokenizer)
        
        # Example on first observation in the training set
        debug.trace_fmt(4, "Sentence: {s}", s=train_InputExamples.iloc[0].text_a)
        debug.trace_fmt(4, "-"*30)
        debug.trace_fmt(4, "Tokens: {t}", t=self.tokenizer.tokenize(train_InputExamples.iloc[0].text_a))
        debug.trace_fmt(4, "-"*30)
        debug.trace_fmt(4, "Input IDs: {i}", i=train_features[0].input_ids)
        debug.trace_fmt(4, "-"*30)
        debug.trace_fmt(4, "Input Masks: {m}", m=train_features[0].input_mask)
        debug.trace_fmt(4, "-"*30)
        debug.trace_fmt(4, "Segment IDs: {i}", i=train_features[0].segment_ids)

        # Initializing the model and the estimator
        model_fn = self.model_fn_builder(
            num_labels=len(self.LABEL_LIST),
            learning_rate=self.LEARNING_RATE,
            num_train_steps=num_train_steps,
            num_warmup_steps=num_warmup_steps)

        self.estimator = tf.estimator.Estimator(
            model_fn=model_fn,
            config=run_config,
            params={"batch_size": self.BATCH_SIZE})

        # Create an input function for training. Use drop_remainder=True when using TPUs.
        train_input_fn = bert.run_classifier.input_fn_builder(
            features=train_features,
            seq_length=self.MAX_SEQ_LENGTH,
            is_training=True,
            drop_remainder=False)

        # Create an input function for validating. Use drop_remainder=True when using TPUs.
        val_input_fn = run_classifier.input_fn_builder(
            features=val_features,
            seq_length=self.MAX_SEQ_LENGTH,
            is_training=False,
            drop_remainder=False)

        # Train the model
        debug.trace(4, "Beginning Training!")
        current_time = datetime.now()
        self.estimator.train(input_fn=train_input_fn, max_steps=num_train_steps)
        debug.trace_fmt(4, "Training took time {t}", t=(datetime.now() - current_time))
        
        # Evaluating the model with Validation set
        self.estimator.evaluate(input_fn=val_input_fn, steps=None)

        # Evaluate over test set
        # note: creates Excel submission file
        pred_sentences = list(test['STORY'])
        predictions = self.getPrediction(pred_sentences)
        debug.trace_expr(4, predictions[0])
        #
        enc_labels = []
        act_labels = []
        for i in range(len(predictions)):
            enc_labels.append(predictions[i][2])
            act_labels.append(predictions[i][3])
        #
        data_file = system.form_path(INPUT_DIR, SUBMISSION_FILE)
        pd.DataFrame(enc_labels, columns=['SECTION']).to_excel(data_file,
                                                               index=False)

        # Classifying random sentences
        tests = self.getPrediction(
            ['Mr.Modi is the Indian Prime Minister',
             'Gaming machines are powered by efficient micro processores and GPUs',
             'That HBO TV series is really good',
             'A trillion dollar economy '
            ])
        debug.trace_object(4, tests, "tests")
        return

#-------------------------------------------------------------------------------
    
if __name__ == '__main__':
    debug.trace_current_context(level=debug.QUITE_DETAILED)
    app = Script(
        description=__doc__,
        skip_input=False,
        manual_input=True,
    )
    app.run()



#! /usr/bin/env python
#
# Takes an input table and transposes the rows and columns. In addition, there
# is support for showing each field value on a separate line prefixed by the
# column name. This is intended to make large table dumps more comprehensible,
# such as from databases.
#------------------------------------------------------------------------
# Sample Input:
#
#   i_timestamp   | i_ip_addr1 |             i_session_id                 | i_keyword
#   1384983367.79 | 1138872328 | 003a4a80db5eda5fa5e7359d57afc29ac1fec377 | Staples Retail Office Products
#   1384983366.04 | 1158147302 | 003b7091f121e03a4ca4e6f8b30e052f78fba19f | Quality
#   1384983366.04 | 1158147302 | 003b7091f121e03a4ca4e6f8b30e052f78fba19f | Quality
#   1384948918.84 | 1130098581 | 003bb1e9a137f6cf1ddd58941c6c7a326c9b2c3d | medical assistant
#
# Sample output:
#
#   i_timestamp: 1384983367.79 | 1384983366.04 | 1384983366.04 | 1384948918.84
#   i_ip_addr1: 1138872328 | 1158147302 | . | 1130098581
#   i_session_id: 003a4a80db5eda5fa5e7359d57afc29ac1fec377 | 003b7091f121e03a4ca4e6f8b30e052f78fba19f | . | 003bb1e9a137f6cf1ddd58941c6c7a326c9b2c3d
#   i_keyword: Staples Retail Office Products | Quality | . | medical assistant
#
#   via: ./transpose-data.py --elide --delim=' | ' < sample-transpose-input.data 
#    
# TODO:
# - Have option to disable use of labels alltogether.
# - Have option to prefix values with column number.
# - Ignore (first) header line if same as headers option.
# - Have separate option for output delim (eg., ": ").
# - Rework field extraction using csv package (rather than split).
# - Have option for setting delim to tab to avoid awkward spec under bash (e.g., --delim $'\t').
#

"""Takes an input table and transposes the rows and columns"""

# Standard packages
import csv
import sys
import argparse

# Local packages
## TODO: from mezcla import debug
## OLD: from tpo_common import debug_print, trace_array, print_stderr
from mezcla.tpo_common import debug_print, trace_array
from mezcla.system import print_stderr
## OLD: from glue_helpers import read_lines
from mezcla.glue_helpers import read_lines

def main():
    """Entry point for script"""
    # Check command-line arguments
    parser = argparse.ArgumentParser(description="Transpose a table from input")
    parser.add_argument("--header", help="File with field names")
    parser.add_argument("--delim", help="Delimiter for fields")
    parser.add_argument("--csv", action='store_true', default=False, help="Comma separated value input")
    parser.add_argument("--encode-newlines", action='store_true', default=False, help="Encode embedded newlines as <EOL>")
    parser.add_argument("--dialect", help="CVS dialect (e.g., excel or unix)")
    parser.add_argument("--elide", dest='elide_fields', action='store_true', default=False, help="Replace repeated values by .'s")
    parser.add_argument("--elided-value", help="Value for repeated field")
    parser.add_argument("--single-field", dest='single_field', action='store_true', default=False, help="Only show a single field per output line")
    parser.add_argument("filename", nargs='?', default='-')
    args = vars(parser.parse_args())
    debug_print("args = %s" % args, 5)
    delim = "\t"
    elided_value = "."
    field_names = []
    field_data = []
    single_field = args['single_field']
    elide_fields = args['elide_fields']
    encode_newlines = args['encode_newlines']
    csv_dialect = args['dialect']
    if args['delim']:
        delim = args['delim']
    if args['csv']:
        delim = ","
    if args['elided_value']:
        elided_value = args['elided_value']
    if args['header']:
        lines = read_lines(args['header'])
        ## OLD: field_names = [label.strip() for label in lines[0].split(delim)]
        header_reader = csv.reader(iter(lines), delimiter=delim, quotechar='"', dialect=csv_dialect)
        field_names = header_reader[0]
        trace_array(field_names, 5, "field_names")
        if not single_field:
            field_data = [[] for i in range(len(field_names))]
            trace_array(field_data, 5, "field_data")
        previous_value = [None] * len(field_names)
    input_stream = sys.stdin
    if (args['filename'] and (args['filename'] != "-")):
        input_stream = open(args['filename'])

    # Transpose each line of the table
    num_lines = 0
    csv_reader = csv.reader(iter(input_stream.readlines()), delimiter=delim, quotechar='"', dialect=csv_dialect)
    ## OLD: for line in input_stream:
    for line_data in csv_reader:
        num_lines += 1
        ## OLD:
        ## line = line.strip("\n")
        ## debug_print("L%d: %s" % (num_lines, line), 6)
        ## line_data = [field.strip() for field in line.split(delim)]
        debug_print("R%d: %s" % (num_lines, line_data), 6)
        trace_array(line_data, 5, "line_data")

        # Use first line as field names if not yet defined
        if (len(field_names) == 0):
            field_names = line_data
            if not single_field:
                field_data = [[] for i in range(len(field_names))]
                trace_array(field_data, 5, "field_data")
            previous_value = [None] * len(field_names)
            continue
        ## OLD: elif ((num_lines == 1) and (field_names == line_data)):
        if ((num_lines == 1) and (field_names == line_data)):
            debug_print("Ignoring duplicate header", 5)
            continue

        # Append each field to respective list (of seen values)
        if (len(line_data) != len(field_names)):
            print_stderr("Warning: Found %d fields but expected %d" % (len(line_data), len(field_names)))
            line_data += (['n/a'] * max(0, len(field_names) - len(line_data)))
        for i in range(len(field_names)):
            debug_print("d[%d]: %s" % (i, line_data[i]), 7)
            new_value = line_data[i]
            if encode_newlines:
                new_value = new_value.replace("\n", "<EOL>")
            if "\n" in new_value:
                new_value = '"' + new_value + '"'
            if (elide_fields and (previous_value[i] == line_data[i])):
                new_value = elided_value
            ## OLD: if (single_field):
            if single_field:
                print("%s" % (delim.join([field_names[i], new_value])))
            else:
                field_data[i].append(new_value)
                previous_value[i] = line_data[i]
        if not single_field:
            trace_array(field_data, 8, "field_data")

    # Output the transposed lines
    if not single_field:
        for i in range(len(field_names)):
            print("%s" % delim.join([field_names[i]] + field_data[i]))

    return

#------------------------------------------------------------------------

if __name__ == '__main__':
    main()



#! /usr/bin/env python
#
# Test(s) for ../main.py
#
# Notes:
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_main.py
# For tips on pytest monkeypatch, see
#   https://stackoverflow.com/questions/38723140/i-want-to-use-stdin-in-a-pytest-test
#

"""Unit tests for main module"""

# Standard packages
from argparse import ArgumentParser
import io
import sys

# Installed packages
import pytest

# Local packages
from mezcla import debug
from mezcla import tpo_common as tpo
from mezcla.unittest_wrapper import TestWrapper

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
import mezcla.main as THE_MODULE

class MyArgumentParser(ArgumentParser):
    """Version of ArgumentParser that doesn't exit upon failure"""

    def __init__(self, *args, **kwargs):
        """Constructor (relegating all to parent)"""
        super().__init__(*args, **kwargs)

    def exit(self, status=0, message=None):
        """Version of exit that doesn't really exit"""
        if message:
            self._print_message(message, sys.stderr)


class TestMain(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.derive_tested_module_name(__file__)
    ## HACK: globals for use in embedded classes (TODO: move into Test class below)
    input_processed = None
    main_step_invoked = None
    process_line_count = -1

    def test_script_options(self):
        """Makes sure script option specifications are parsed OK"""
        debug.trace(4, f"in test_script_options(); self={self}")
        class Test(THE_MODULE.Main):
            """"Dummy test class"""
            argument_parser = MyArgumentParser
            ## TODO: rename as TestMain?

        # note: format is ("option", "description", "default"), or just "option"
        app = Test(text_options=[("name", "Full name", "John Doe")],
                   boolean_options=[("verbose", "testing verbose option", True)],
                   )
        #
        assert app.parsed_args.get("name") == "John Doe"
        assert app.parsed_args.get("verbose")

    @pytest.mark.xfail
    def test_script_without_input(self):
        """Makes sure script class without input doesn't process input and that
        the main step gets invoked"""
        debug.trace(4, f"in test_script_without_input(); self={self}")

        # This avoids flaky stderr due to other tests
        tpo.restore_stderr()

        # Create scriptlet checking for input and processing main step
        # TODO: rework with external script as argparse exits upon failure
        class Test(THE_MODULE.Main):
            """"Dummy test class"""
            argument_parser = MyArgumentParser

            def setup(self):
                """Post-argument parsing processing: just displays context"""
                tpo.debug_format("in setup(): self={s}", 5, s=self)
                TestMain.process_line_count = 0
                tpo.trace_current_context(label="Test.setup", 
                                          level=tpo.QUITE_DETAILED)
                tpo.trace_object(self, tpo.QUITE_DETAILED, "Test instance")
            #
            def process_line(self, line):
                """Dummy input processing"""
                tpo.debug_format("in Test.process_line({l}): self={s}", 5,
                                 l=line, s=self)
                TestMain.input_processed = True
                TestMain.process_line_count += 1

            #
            def run_main_step(self):
                """Dummy main step"""
                tpo.debug_format("in Test.run_main_step()): self={s}", 5,
                                 s=self)
                TestMain.main_step_invoked = True

        # Test scriptlet with test script as input w/ and w/o input enabled
        TestMain.input_processed = None
        app1 = Test(skip_input=False, runtime_args=[__file__])
        try:
            app1.run()
        except:
            tpo.print_stderr("Exception during run method: {exc}",
                             exc=tpo.to_string(sys.exc_info()))
        assert TestMain.input_processed
        #
        TestMain.input_processed = None
        app2 = Test(skip_input=True, runtime_args=[__file__])
        try:
            app2.run()
        except:
            tpo.print_stderr("Exception during run method: {exc}",
                             exc=tpo.to_string(sys.exc_info()))
        assert not TestMain.input_processed

        # Test scriptlet w/ input disabled and wihout arguments
        # note: auto_help disabled so that no arguments needed
        TestMain.main_step_invoked = None
        app3 = Test(skip_input=True, manual_input=True, auto_help=False, runtime_args=[])
        try:
            app3.run()
        except:
            tpo.print_stderr("Exception during run method: {exc}",
                             exc=tpo.to_string(sys.exc_info()))
        assert TestMain.main_step_invoked

    @pytest.mark.xfail
    def test_perl_arg(self):
        """Make sure perl-style arg can be parsed"""
        # TODO: create generic app-creation helper
        debug.trace(4, f"in test_perl_arg(); self={self}")
        class Test(THE_MODULE.Main):
            """"Dummy test class"""
            argument_parser = MyArgumentParser

        # Test with and without Perl support
        app = Test(boolean_options=[("verbose", "testing verbose option")],
                   perl_switch_parsing=True)
        #
        assert app.parsed_args.get("verbose") == 0
        #
        app = Test(boolean_options=[("verbose", "testing verbose option")],
                   perl_switch_parsing=False)
        # NOTE: this ensures that is None and not 0
        assert app.parsed_args.get("verbose") is None


class TestMain2:
    """Another class for testcase definition
    Note: Needed to avoid error with pytest due to inheritance with unittest.TestCase via TestWrapper"""

    def test_input_modes(self, capsys, monkeypatch):
        """Make sure input processed OK with respect to line/para/file mode"""
        debug.trace(4, f"in test_input_modes({capsys}); self={self}")
        contents = "1\n\n2\n\n\n3\n\n\n\n4\n\n\n\n"
        num_lines = len(contents.split("\n"))
        pre_captured = capsys.readouterr()
        # line input
        monkeypatch.setattr('sys.stdin', io.StringIO(contents))
        main = THE_MODULE.Main(skip_args=True, auto_help=False)
        main.run()
        captured = capsys.readouterr()
        ## TODO: assert(TestMain.process_line_count == num_lines)
        assert num_lines == len(captured.out.split("\n"))
        debug.trace_expr(5, main, num_lines)
        # paragraph input
        monkeypatch.setattr('sys.stdin', io.StringIO(contents))
        main = THE_MODULE.Main(paragraph_mode=True, skip_args=True, auto_help=False)
        main.run()
        captured = capsys.readouterr()
        ## TODO: assert(TestMain.process_line_count == 4)
        assert(num_lines == len(captured.out.split("\n")))
        debug.trace_expr(5, main, num_lines)
        # file input
        monkeypatch.setattr('sys.stdin', io.StringIO(contents))
        main = THE_MODULE.Main(file_input_mode=True, skip_args=True, auto_help=False)
        main.run()
        captured = capsys.readouterr()
        ## TODO: assert(TestMain.process_line_count == 1)
        assert(num_lines == len(captured.out.split("\n")))
        debug.trace_expr(5, main, num_lines, pre_captured)

    def test_missing_newline(self, capsys, monkeypatch):
        """Make sure file with missing newline at end processed OK"""
        debug.trace(4, f"in test_missing_newline({capsys}); self={self}")
        contents = "1\n2\n3"
        num_lines = len(contents.split("\n"))
        _pre_captured = capsys.readouterr()
        monkeypatch.setattr('sys.stdin', io.StringIO(contents))
        main = THE_MODULE.Main(skip_args=True, auto_help=False)
        main.run()
        captured = capsys.readouterr()
        # note: 1 extra line ('') and 1 extra character (final newline)
        assert (1 + num_lines) == len(captured.out.split("\n"))
        assert (1 + len(contents)) == len(captured.out)
        ## TODO: assert TestMain.process_line_count == 3
        debug.trace_expr(5, main, num_lines)

    def test_has_parsed_option_hack(self):
        """Make sure (temporarily hacked) has_parsed_option differs from has_parsed_option_old"""
        debug.trace(4, f"in test_has_parsed_option_hack(); self={self}")
        ok_arg = "ok"
        missing_arg = "missing"
        main = THE_MODULE.Main(skip_args=True, auto_help=False)
        main.parsed_args = {ok_arg: True}
        assert main.has_parsed_option_old(ok_arg) == main.has_parsed_option(ok_arg)
        assert main.has_parsed_option_old(missing_arg) != main.has_parsed_option(missing_arg)

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    pytest.main([__file__])



#! /usr/bin/env python
#
# format-profile.py: displays the result of python profiling via cProfile
#
# usage example:
#    python -u -m cProfile -o fubar.profile fubar.py
#    format-profile.py fubar.profile >| fubar.profile.list
#------------------------------------------------------------------------
# Notes:
# via http://docs.python.org/3/library/profile.html:
#
# The Stats Class
#
# Analysis of the profiler data is done using the Stats class.
#
# class pstats.Stats(*filenames or profile, stream=sys.stdout)
# ...
# Stats objects have the following methods:
# ...
# 
# sort_stats(*keys)
#
# This method modifies the Stats object by sorting it according to the
# supplied criteria. The argument is typically a string identifying
# the basis of a sort (example: 'time' or 'name').
#
# Valid Arg    Meaning
# 'calls'      call count
# 'cumulative' cumulative time
# 'cumtime'    cumulative time
# 'file'       file name
# 'filename'   file name
# 'module'     file name
# 'ncalls'     call count
# 'pcalls'     primitive call count
# 'line'       line number
# 'name'       function name
# 'nfl'        name/file/line
# 'stdname'    standard name
# 'time'       internal time
# 'tottime'    internal time
#
# Note: following apply to version 2.7 or later
# - cumtime, filename, ncalls, tottime
#

"""Displays result of python profiling"""

#------------------------------------------------------------------------
# Library packages

# Standard packages
import sys
import pstats

# Local packages
from mezcla import debug
## OLD: from tpo_common import *
from mezcla.system import getenv_text, print_stderr

## OLD: PROFILE_KEY = getenv_text("PROFILE_KEY", "cumulative")
PROFILE_KEY = getenv_text("PROFILE_KEY", "cumulative",
                          "Sort key (e.g., calls, cumulative, file, time)")

#------------------------------------------------------------------------
# Functions

def usage():
    """Displays usage notes for script"""
    print_stderr("""

Usage: {program} profile-log

Notes:
- use PROFILE_KEY to over default sorting (cumulative)
- main keys: 
       calls, cumulative, file, time
- other keys: 
       module, pcalls, line, name, nfl, stdname
- 2.7+ keys: 
       cumtime, filename, ncalls, tottime
- unfortunately, memory profiling is not supported
- see http://docs.python.org/3/library/profile.html

Example (assumes bash):
    $ python -m cProfile -o profile.log fubar.py
    $ PROFILE_KEY=calls {program} profile.log

""".format(program=sys.argv[0]))
    return
                

def main():
    """Entry point for script"""
    debug.trace_fmt(5, "main(): args={v}", v=sys.argv)
    if ((len(sys.argv) < 2) or (sys.argv[1] == "--help")):
        usage()
        sys.exit()
    file = sys.argv[1]

    # Generate listing and sort by cumulative time
    p = pstats.Stats(file)
    p.strip_dirs().sort_stats(PROFILE_KEY).print_stats()
    return

#------------------------------------------------------------------------

if __name__ == '__main__':
    main()



#! /usr/bin/env python
#
# Adhoc docstring tests for mezcla.
#
# Notes:
# - The doctest comparison is picky, so you might need to account for whitespace
#   trivial differences.
#

r"""
Miscellaenous docstring tests for mezcla modules

#...............................................................................
# Initialization

## TODO?
## >>> import re

#...............................................................................
# __init__.py

>>> from mezcla import *

>>> int(TL.DEFAULT)
2

>>> int(TL.DETAILED)
4

>>> (len(set(["debug", "mezcla", "system"]).intersection(dir())) >= 3)
True

>>> mezcla_dir = mezcla.__path__[0]

#................................................................................
# main.py

# Make sure default temp_file under $TMP/main-...
# main.dummy_app.temp_file 
# '/home/tomohara/temp/tmp/main-y8c4o3vx'

>>> from mezcla.main import dummy_app

>>> (isinstance(dummy_app.temp_file, str) and (len(dummy_app.temp_file) > 0))
True

>>> re.sub(r"-[^\\-]+$", "-...",
...        re.sub(system.getenv_text('TMP', '???'), "<TMP>",
...               dummy_app.temp_file))
'<TMP>/main-...'

## TEST:
## >>> gh.run("bash -i -c 'BRIEF_USAGE=0 main.py 2>&1'")
## ...

>>> system.setenv("BRIEF_USAGE", "0")

>>> gh.run("main.py 2>&1")
Warning: /home/tomohara/python/Mezcla/mezcla/main.py is not intended to be run standalone
usage: main.py [-h] [filename]

TODO: what the script does

positional arguments:
  filename    Input filename

optional arguments:
  -h, --help  show this help message and exit

#................................................................................
# debug.py

>>> temp_temp_dir = ""

>>> def init_temp_temp_dir():
...     '''adhoc function'''
...     global temp_temp_dir
...     temp_temp_dir = dummy_app.temp_file + "-temp-dir"

>>> debug.code(1, init_temp_temp_dir)

>>> (len(temp_temp_dir) > 0)
True

#................................................................................
# glue_helpers.py
#
# note: shameless hack used below (effing dostests rigidity)
#

>>> gh.full_mkdir(temp_temp_dir)

## TEST:
## >>> gh.run(f"bash -i -c \"cd '{temp_temp_dir}'; touch 1 3 5; ls\"")
##

>>> gh.run(f"cd '{temp_temp_dir}'; touch 1 3 5; ls")
1
3
5

"""

# Standard packages
import doctest
import re
import sys

# Local packages
from mezcla import debug
from mezcla import system

# Constants
TL = debug.TL

#-------------------------------------------------------------------------------

def main():
    """Entry point"""
    debug.trace(TL.USUAL, f"main(): filename={system.real_path(__file__)}")

    # Show usage if insuffient command line
    # TODO: add real argument processing
    command_line = str(sys.argv)
    show_usage = ((len(sys.argv) <= 1)
                  or (re.search(r"--help\b", command_line)))
    if show_usage:
        system.print_error(f"Usage: {sys.argv[0]} [--help] [--verbose | --quiet] [--diff]")
        system.print_error("")
        system.print_error("Note: See test_doctest.py for plenty of example tests:")
        system.print_error("    https://github.com/python/cpython/blob/main/Lib/test/test_doctest.py")
        system.print_error("")
        system.print_error("Examples:\n\n")
        system.print_error(f"{sys.argv[0]} --quiet\n\n")
        system.print_error(f"{sys.argv[0]} --verbose 2>&1 | less\n\n")
        sys.exit()
    debug.trace_expr(4, command_line, show_usage)
                           
    # Run tests, optionally in verbose
    verbose = re.search(r"--verbose\b", command_line)
    debug.trace_expr(4, verbose)
    options = doctest.NORMALIZE_WHITESPACE
    if re.search(r"--diff\b", command_line):
        options |= doctest.REPORT_CDIFF
    debug.trace(5, f"doctest options: {options}; {doctest.OPTIONFLAGS_BY_NAME}")
    doctest.testmod(verbose=verbose,
                    optionflags=options)
                           

if __name__ == "__main__":
    main()



#! /usr/bin/env python
#
# Test(s) for ../template.py
#
# Notes:
# - This is simple test of module not the test template (see ./template.py).
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_template.py
#

"""Tests for template module"""

# Standard packages
import re

# Installed packages
import pytest

# Local packages
from mezcla.unittest_wrapper import TestWrapper
from mezcla import debug
from mezcla import glue_helpers as gh

# Note: Two references are used for the module to be tested:
#    THE_MODULE:	    global module object
## TODO: template => new name
import mezcla.template as THE_MODULE
#
# Note: sanity test for customization (TODO: remove if desired)
if not re.search(__file__, r"\btemplate.py$"):
    debug.assertion("mezcla.template" not in str(THE_MODULE))


class TestTemplate(TestWrapper):
    """Class for testcase definition"""
    script_file = TestWrapper.get_module_file_path(__file__)
    script_module = TestWrapper.get_testing_module_name(__file__)

    def test_data_file(self):
        """Makes sure to-do grep works as expected"""
        debug.trace(4, "TestTemplate.test_data_file()")
        data = ["TEMP", "TODO", "DONE"]
        gh.write_lines(self.temp_file, data)
        output = self.run_script("--TODO-arg", self.temp_file)
        assert re.search(r"arg1 line \(2\): TODO", output.strip())
        return

    def test_something_else(self):
        """Make sure arg value reflects to-do nature"""
        debug.trace(4, "test_something_else()")
        assert "todo" in THE_MODULE.TODO_ARG.lower()
        return

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    pytest.main([__file__])



#! /usr/bin/env python
#
# Based on following:
#   https://www.geeksforgeeks.org/downloading-gists-from-github-made-simple
#
# TODO:
#  - Get the script to work!
#

"""
Download user-specific gists from Github

    Commands:
    show:               To show all the available gists with their assigned gistno.
    download all:       To download all the available gists.
    download gistno(s): To download gist(s) assigned to gistno(s).
    detailed gistno:    To print content of gist assigned to gistno.
    exit:               To exit the script.

"""

# Standard modules
import os

# Intalled module
import requests

# Local modules
from mezcla import debug
from mezcla.main import Main
from mezcla import glue_helpers as gh
from mezcla import system


def create_directory(dirname):
    """Creates a new directory if a directory with dirname does not exist"""
    debug.trace(4, f"create_directory({dirname})")

    try:
        os.stat(dirname)
    except:
        os.mkdir(dirname)


def show(obj):
    """Displays the items in the obj"""
    debug.trace(4, f"show({obj})")

    for i in range(len(obj)):
        print(str(i)+': '+str(obj[i]))


def auth():
    """Asks for the user details"""
    debug.trace(4, "auth()")

    ask_auth = input("Do you want to download gists from your account"
                     "? Type 'yes' or 'no': ")
    debug.trace(6, ask_auth)
    user = None
    request = None
    if (ask_auth=="yes"):
        user = input("Enter your username: ")
        password = input("Enter your password: ")
        debug.trace_expr(3, user, password)
        request = requests.get('https://gists.github.com/users/' + user,
                               auth=(user, password))
    elif (ask_auth=="no"):
        user = input("Enter username: ")
        debug.trace_expr(3, user)
        ## OLD
        ## request = requests.get('https://api.github.com/users/'
        ##                        +user+'/gists')
        request = requests.get('https://gists.github.com/' + user)
        debug.trace_object(5, request)
    else:
        system.print_error("Error: unexpected condition in auth: {ask_auth}")
    debug.trace(5, f"auth() => {[ask_auth, user, request]}")
    return [ask_auth, user, request]


def load(request):
    """Loads the files and the gist urls"""
    debug.trace(4, f"load({request})")

    output = request.text.split(",")
    debug.trace_expr(6, output)
    gist_urls = []
    files = []
    for item in output:
        debug.trace_expr(6, item)
        if "raw_url" in item:
            gist_urls.append(str(item[11:-1]))
        if "filename" in item:
            files.append(str(item.split(":")[1][2:-1]))
    debug.trace(5, f"load() => {[gist_urls, files]}")
    
    return [gist_urls, files]


def write_gist(filename, text):
    """Writes text(gist) to filename"""
    debug.trace(4, f"write_gist({filename}, {gh.elide(text)})")

    fp = open(filename, mode='w', encoding='utf-8')
    fp.write(text)
    fp.close()


def download(permission, user, request, fileno):
    """Loads and writes all the gists to <em>dirname</em>"""
    debug.trace(4, f"download({permission}, {user}, {request}, {fileno})")

    ## OLD: if(permission == "yes" or permission == "no"):
    if permission in ["yes", "no"]:
        gist_urls, files = load(request)
        debug.trace_expr(6, gist_urls, files)
        ## OLD: dirname = user+"'s_gists/"
        dirname = user + "_gists/"
        create_directory(dirname)
        if(fileno[1] == "all"):
            for i in range(len(gist_urls)):
                gist = requests.get(gist_urls[i])
                write_gist(dirname+files[i], gist.text)
        else:
            for i in range(1,len(fileno)):
                gist = requests.get(gist_urls[int(fileno[i])])
                write_gist(dirname+files[int(fileno[i])], gist.text)


def detailed(urls, pos):
    """Prints out the contents of a file"""
    debug.trace(4, f"detailed({urls}, {pos})")

    gist = requests.get(urls[int(pos)])
    print(gist.text)


def main():
    """Authenticates and downloads gists according to user's choice"""
    debug.trace(4, "main()")
    
    # Show simple usage if --help given
    dummy_app = Main(description=__doc__, skip_input=False, manual_input=False)

    ask_auth, user, request = auth()
    urls, files = load(request)
    try:
        while(1):
            command = input("Enter your command: ").strip()
            debug.trace_expr(3, command)
            if ("download" in command):
                download(ask_auth, user, request, command.split(" "))
            elif ("detailed" in command):
                detailed(urls, command.split(" ")[1])
            elif (command == "show"):
                show(files)
            elif (command == "exit"):
                return
            elif (command == ""):
                pass
            else:
                system.print_error(f"Unknown command '{command}'\nUsage follows:{__doc__}")
    except:
        pass

if(__name__ == '__main__'):
    main()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
# Convert emoticons into names (or just strips them)
#
# Example Input:
#   Nothing to do 😴
#
# Example output:
#   Nothing to do [sleeping face]
#

"""
Replace emoticons with name (or remove entirely)

Sample usage:

   echo 'github-\U0001F634_Transformers' | {script} -
"""                          # 🤗 (HuggingFace logo)

# Standard modules
import unicodedata

# Intalled module
# TODO: import numpy as np

# Local modules
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.main import Main
## TODO: from mezcla.my_regex import my_re
from mezcla import system
## TODO2: streamline imports by exposing common functions, etc. in mezcla

# Constants
TL = debug.TL
STRIP_OPT = "strip"

## TODO:
## # Environment options
## # Notes:
## # - These are just intended for internal options, not for end users.
## # - They also allow for enabling options in one place rather than four
## #   when using main.Main (e.g., [Main member] initialization, run-time
## #   value, and argument spec., along with string constant definition).
## #
## ENABLE_FUBAR = system.getenv_bool("ENABLE_FUBAR", False,
##                                   description="Enable fouled up beyond all recognition processing")

#-------------------------------------------------------------------------------

OTHER_SYMBOL = 'So'
#
def convert_emoticons(text, replace=None, strip=None):
    """Either REPLACE emotions with Unicode name or STRIP them entirely"""
    # EX: convert_emoticons("✅  Success") => "[checkmark] Success"
    # EX: convert_emoticons("¿Hablas español?") => "¿Hablas español?" # Spanish for "Do you speak Spanish"
    # EX: convert_emoticons("天気") => "天気"   # Japanese for weather
    debug.trace(6, f"convert_emoticons(_, [r={replace}], [s={strip}])")
    debug.assertion(not (replace and strip))
    if replace is None:
        replace = not strip
    in_text = text
    chars = []
    for ch in text:
        if unicodedata.category(ch) == OTHER_SYMBOL:
            ch = f"[{unicodedata.name(ch).lower()}]" if replace else ""
        chars.append(ch)
    text = "".join(chars)
    level = (4 if (text != in_text) else 6)
    debug.trace(level, f"convert_emoticons({in_text}) => {text}")
    return text


def main():
    """Entry point"""
    debug.trace(TL.USUAL, f"main(): script={system.real_path(__file__)}")

    # Parse command line options, show usage if --help given
    # TODO: manual_input=True; short_options=True
    main_app = Main(description=__doc__.format(script=gh.basename(__file__)),
                    boolean_options=[(STRIP_OPT, "Strip emoticon entirely, instead of replacing with name")],
                    skip_input=False)
    debug.assertion(main_app.parsed_args)
    strip_entirely = main_app.get_parsed_option(STRIP_OPT)

    for line in main_app.read_entire_input().splitlines():
        print(convert_emoticons(line, strip=strip_entirely))
    return

#-------------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context(level=TL.QUITE_VERBOSE)
    main()



#! /usr/bin/env python
#
# Test(s) for Hugging Face (HF) Stable Diffulion (SD) module: ../hf_stable_diffusion.py
#
# Notes:
# - Fill out TODO's below. Use numbered tests to order (e.g., test_1_usage).
# - TODO: If any of the setup/cleanup methods defined, make sure to invoke base
#   (see examples below for setUp and tearDown).
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by  unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_hf_stable_diffusion.py
#
# TODO3: check image attributes (e.g., backgfround color) and work in image classification (e.g., box)
#

"""Tests for hf_stable_diffusion module"""

# Standard packages
import base64

# Installed packages
import pytest
try:
    import diffusers
except:
    diffusers = None
try:
    import extcolors
except:
    extcolors = None

# Local packages
from mezcla.unittest_wrapper import TestWrapper
from mezcla import debug
from mezcla import system
from mezcla import glue_helpers as gh
from mezcla.my_regex import my_re

# Note: Two references are used for the module to be tested:
#    THE_MODULE:                  global module object
#    TestTemplate.script_module:  path to file
import mezcla.examples.hf_stable_diffusion as THE_MODULE
#
# Note: sanity test for customization (TODO: remove if desired)
if not my_re.search(__file__, r"\btemplate.py$"):
    debug.assertion("mezcla.template" not in str(THE_MODULE))

class TestIt(TestWrapper):
    """Class for testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    # -or- non-mezcla: script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)
    #
    use_temp_base_dir = True            # treat TEMP_BASE as directory
    # note: temp_file defined by parent (along with script_module, temp_base, and test_num)

    @pytest.mark.skipif(not diffusers, reason="SD diffusers package missing")
    def test_simple_generation(self):
        """Makes sure simple image generation works as expected"""
        debug.trace(4, f"TestIt.test_data_file(); self={self}")
        # ex: See sd-app-image-1.png
        output = self.run_script(
            options="--batch --prompt 'orange ball' --negative 'green blue red yellow purple pink brown'",
            env_options=f"BASENAME='{self.temp_base}' LOW_MEMORY=1", uses_stdin=True)
        debug.trace_expr(5, output)
        assert (my_re.search(r"See (\S+.png) for output image\(s\).", output.strip()))
        image_file = my_re.group(1)
        # ex: sd-app-image-3.png: PNG image data, 512 x 512, 8-bit/color RGB, non-interlaced
        file_info = gh.run(f"file {image_file}")
        debug.trace_expr(5, file_info)
        assert (my_re.search("PNG image data, 512 x 512, 8-bit/color RGB", file_info))
        # TODO2: orange in rgb-color profile for image
        return

#...............................................................................

class TestIt2:
    """Another class for testcase definition
    Note: Needed to avoid error with pytest due to inheritance with unittest.TestCase via TestWrapper"""
    
    @pytest.mark.skipif(not diffusers, reason="SD diffusers package missing")
    def test_pipeline(self):
        """Make sure valid SD pipeline created"""
        debug.trace(4, f"TestIt2.test_something_else(); self={self}")
        sd = THE_MODULE.StableDiffusion(use_hf_api=True)
        pipe = sd.init_pipeline()
        actual = my_re.split(r"\W+", str(pipe))
        expect = "CLIPImageProcessor CLIPTextModel StableDiffusionSafetyChecker text_encoder".split()
        debug.trace_expr(5, actual, expect, delim="\n")
        assert (len(system.intersection(actual, expect)) > 2)
        return

    
    @pytest.mark.xfail                   # TODO: remove xfail
    @pytest.mark.skipif(not extcolors, reason="extcolors package missing")
    def test_something_else(self):
        """Make sure prompted color is used"""
        debug.trace(4, f"TestIt2.test_something_else(); self={self}")
        sd = THE_MODULE.StableDiffusion(use_hf_api=True, low_memory=True)
        images = sd.infer(prompt="a ripe orange", scale=30)
        # note: encodes image base-64 str data into bytes and then decodes into image bytes
        image_data = (base64.decodebytes(images[0].encode()))
        image_path = gh.create_temp_file(image_data, binary=True)
        # note: use of rgb_color_name.py allows for fudge factor
        # $ extcolors sd-app-image-1.png | rgb_color_name.py - | grep orange
        # <(255, 92, 0), orangered>   :  47.07% (123388)
        # <(255, 153, 0), orange>  :   6.13% (16074)
        output = gh.run(f"extcolors '{image_path}' | rgb_color_name.py - 2> /dev/null")
        debug.trace_expr(4, output)
        assert ("orange" in output)
        assert (len(images) == 1)
        return

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    pytest.main([__file__])
